#!/usr/bin/env python3
"""
Remap a font's encoded glyphs.

Changes font cmap table. User can also specify their
own output path.

Usage:

gftools remap-font --deep -o font-remap.ttf font.ttf A=A.alt B=B.alt U+0175=wcircumflex.alt
gftools remap-font --map-file=remapping.txt -o font-remap.ttf font.ttf
"""
import argparse
import sys

from fontTools.ttLib import TTFont


def grovel_substitutions(font, lookup, glyphmap):
    if lookup.LookupType == 7:
        raise NotImplementedError
    gmap = lambda g: glyphmap.get(g, g)
    go = font.getGlyphOrder()

    def do_coverage(c):
        c.glyphs = list(sorted([gmap(g) for g in c.glyphs], key=lambda g: go.index(g)))
        return c

    for st in lookup.SubTable:
        if lookup.LookupType == 1:
            newmap = {}
            for inglyph, outglyph in st.mapping.items():
                newmap[gmap(inglyph)] = gmap(outglyph)
            st.mapping = newmap
        elif lookup.LookupType == 2:
            newmap = {}
            for inglyph, outglyphs in st.mapping.items():
                newmap[gmap(inglyph)] = [gmap(g) for g in outglyphs]
            st.mapping = newmap
        elif lookup.LookupType == 4:
            newligatures = {}
            for outglyph, inglyphs in st.ligatures.items():
                for ig in inglyphs:
                    ig.LigGlyph = gmap(ig.LigGlyph)
                    ig.Component = [gmap(c) for c in ig.Component]
                newligatures[gmap(outglyph)] = inglyphs
            st.ligatures = newligatures
        elif lookup.LookupType == 5:
            if st.Format == 1:
                do_coverage(st.Coverage)
                for srs in st.SubRuleSet:
                    for subrule in srs.SubRule:
                        subrule.Input = [gmap(c) for c in subrule.Input]
            elif st.Format == 2:
                do_coverage(st.Coverage)
                st.ClassDef.classDefs = {
                    gmap(k): v for k, v in st.ClassDef.classDefs.items()
                }
            else:
                st.Coverage = [do_coverage(c) for c in st.Coverage]
        elif lookup.LookupType == 6:
            if st.Format == 1:
                do_coverage(st.Coverage)
                for srs in st.ChainSubRuleSet:
                    for subrule in srs.ChainSubRule:
                        subrule.Backtrack = [gmap(c) for c in subrule.Backtrack]
                        subrule.Input = [gmap(c) for c in subrule.Input]
                        subrule.LookAhead = [gmap(c) for c in subrule.LookAhead]
            elif st.Format == 2:
                do_coverage(st.Coverage)
                st.BacktrackClassDef.classDefs = {
                    gmap(k): v for k, v in st.BacktrackClassDef.classDefs.items()
                }
                st.InputClassDef.classDefs = {
                    gmap(k): v for k, v in st.InputClassDef.classDefs.items()
                }
                st.LookAheadClassDef.classDefs = {
                    gmap(k): v for k, v in st.LookAheadClassDef.classDefs.items()
                }
            elif st.Format == 3:
                st.BacktrackCoverage = [do_coverage(c) for c in st.BacktrackCoverage]
                st.InputCoverage = [do_coverage(c) for c in st.InputCoverage]
                st.LookAheadCoverage = [do_coverage(c) for c in st.LookAheadCoverage]


def main(args=None):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--map-file", metavar="TXT", help="Newline-separated mappings")
    parser.add_argument("--output", "-o", metavar="TTF", help="Output font binary")
    parser.add_argument(
        "--deep", action="store_true", help="Also remap inside GSUB table"
    )
    parser.add_argument("font", metavar="TTF", help="Input font binary")
    parser.add_argument("mapping", nargs="*", help="Codepoint-to-glyph mapping")

    args = parser.parse_args(args)

    if not args.mapping and not args.map_file:
        print("You must either specify a mapping or a map file")
        sys.exit(1)
    if args.mapping and args.map_file:
        print("You must specify either a mapping or a map file, not both")
        sys.exit(1)

    mapping = {}
    font = TTFont(args.font)

    if args.map_file:
        incoming_map = open(args.map_file).readlines()
    else:
        incoming_map = args.mapping

    glyph_mapping = {}  # Map glyphname->glyphname
    cmap = font.getBestCmap()
    reversed_cmap = font["cmap"].buildReversed()

    for entry in incoming_map:
        entry = entry.strip()
        if not entry or entry.startswith("#"):
            continue
        codepoint, newglyph = entry.split("=")
        if codepoint.startswith("U+") or codepoint.startswith("0x"):
            codepoint = int(codepoint[2:], 16)
        elif codepoint in cmap:
            codepoint = cmap[codepoint]
        elif args.deep:  # It's a glyph name?
            oldglyph = codepoint
            if oldglyph not in font.getGlyphOrder():
                print(f"Glyph '{oldglyph}' not found in font")
                continue
            if newglyph not in font.getGlyphOrder():
                print(f"Glyph '{newglyph}' not found in font")
                continue
            glyph_mapping[oldglyph] = newglyph
            # Fix up cmap if we can
            for old_codepoint in reversed_cmap.get(oldglyph, []):
                mapping[old_codepoint] = newglyph
            continue
        mapping[codepoint] = newglyph
        if newglyph not in font.getGlyphOrder():
            print(
                f"Glyph '{newglyph}' (to be mapped to U+{codepoint:04X}) not found in font"
            )
            sys.exit(1)
        if codepoint in cmap:
            glyph_mapping[cmap[codepoint]] = newglyph

    if args.deep:
        for lookup in font["GSUB"].table.LookupList.Lookup:
            grovel_substitutions(font, lookup, glyph_mapping)

    cmap = font["cmap"]
    for table in cmap.tables:
        for codepoint, glyph in mapping.items():
            table.cmap[codepoint] = glyph

    if args.output:
        out = args.output
    else:
        out = args.font
    print(f"Saving font: {out}")
    font.save(out)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Check a font family using Google Fonts QA tools.

Examples:
Check a local family against the same family hosted on Google Fonts:
`gftools qa -f [fonts.ttf] -gfb -a -o qa`

Check a local family against another local family and generate reports
for Font Diffenator only:
`gftools qa -f [fonts_a.ttf] -fb [fonts_b.ttf] --diffenator -o qa`


Compare a pull request against the same family hosted on Google Fonts:
`gftools qa -pr www.github.com/user/repo/pull/1 -gfb -a -o qa`

Compare a github folder of fonts against the same family hosted on Google
Fonts:
`gftools qa -gh www.github.com/user/repo/tree/fonts/ttf -gfb -a -o qa`
"""
from fontTools.ttLib import TTFont
import argparse
import os
import sys
import shutil
import logging
from gftools.utils import (
    download_family_from_Google_Fonts,
    download_files_in_github_pr,
    download_files_in_github_dir,
    download_files_from_archive,
    Google_Fonts_has_family,
    mkdir,
)
import re
from gftools.qa import FontQA
from diffenator2.font import DFont


__version__ = "3.1.0"
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def family_name_from_fonts(fonts):
    results = set(f.family_name for f in fonts)
    if len(results) > 1:
        raise Exception("Multiple family names found: [{}]".format(", ".join(results)))
    return list(results)[0]


def main(args=None):
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    font_group = parser.add_argument_group(title="Fonts to qa")
    font_input_group = font_group.add_mutually_exclusive_group(required=True)
    font_input_group.add_argument("-f", "--fonts", nargs="+", help="Paths to fonts")
    font_input_group.add_argument(
        "-pr", "--pull-request", help="Get fonts from a Github pull request"
    )
    font_input_group.add_argument(
        "-gh", "--github-dir", help="Get fonts from a Github directory"
    )
    font_input_group.add_argument(
        "-gf", "--googlefonts", help="Get fonts from Google Fonts"
    )
    font_input_group.add_argument(
        "-ar", "--archive", help="Get fonts from a zip file URL"
    )

    font_before_group = parser.add_argument_group(title="Fonts before input")
    font_before_input_group = font_before_group.add_mutually_exclusive_group(
        required=False
    )
    font_before_input_group.add_argument(
        "-fb", "--fonts-before", nargs="+", help="Paths to previous fonts"
    )
    font_before_input_group.add_argument(
        "-prb",
        "--pull-request-before",
        help="Get previous fonts from a Github pull request",
    )
    font_before_input_group.add_argument(
        "-ghb", "--github-dir-before", help="Get previous fonts from a Github dir"
    )
    font_before_input_group.add_argument(
        "-arb", "--archive-before", help="Get previous fonts from a zip file URL"
    )
    font_before_input_group.add_argument(
        "-gfb",
        "--googlefonts-before",
        action="store_true",
        help="Get previous fonts from Google Fonts",
    )

    check_group = parser.add_argument_group(title="QA checks")
    check_group.add_argument(
        "-a",
        "--auto-qa",
        action="store_true",
        help="Check fonts against against the same fonts hosted on Google Fonts",
    )
    check_group.add_argument(
        "--diffenator", action="store_true", help="Run Fontdiffenator"
    )
    check_group.add_argument("--proof", action="store_true", help="Run HTML proofs")
    check_group.add_argument(
        "--render",
        action="store_true",
        help="Run diffbrowsers if fonts_before exist, otherwise run proof",
    )
    check_group.add_argument("--fontbakery", action="store_true", help="Run FontBakery")
    check_group.add_argument(
        "--diffbrowsers", action="store_true", help="Run Diffbrowsers"
    )
    check_group.add_argument(
        "--interpolations", action="store_true", help="Run interpolation checker"
    )
    parser.add_argument("-re", "--filter-fonts", help="Filter fonts by regex")
    parser.add_argument(
        "-o", "--out", default="out", help="Output path for check results"
    )
    parser.add_argument(
        "-ogh",
        "--out-github",
        action="store_true",
        help=(
            "Post report data to either the pull request as a comment "
            "open a new issue. This can only be used if fonts have been "
            "fetched from either a pull request or github dir."
        ),
    )
    parser.add_argument(
        "--out-url",
        help=(
            "Post report data to a github pr. This can be used with any font "
            "fetching method."
        ),
    )
    parser.add_argument("--rust", action="store_true", help="Use Rust tooling")
    check_group.add_argument(
        "--extra-fontbakery-args",
        help="Additional arguments to Fontbakery",
        action="append",
    )
    check_group.add_argument(
        "--extra-fontspector-args",
        help="Additional arguments to Fontspector",
        action="append",
    )

    parser.add_argument("--imgs", action="store_true", help="Gen images using Selenium")
    parser.add_argument("--version", action="version", version=__version__)
    args = parser.parse_args(args)
    if args.out_github and not any([args.pull_request, args.github_dir]):
        raise Exception(
            "Cannot upload results to a github issue or pr. "
            "Font input must either a github dir or a pull request"
        )
    if not any(
        [
            args.auto_qa,
            args.fontbakery,
            args.proof,
            args.diffbrowsers,
            args.diffenator,
            args.render,
            args.interpolations,
        ]
    ):
        raise Exception(
            "Terminating. No checks selected. Run gftools qa "
            "--help to see all possible commands."
        )

    # Retrieve fonts and store in out dir
    mkdir(args.out)
    fonts_dir = os.path.join(args.out, "fonts")
    mkdir(fonts_dir)
    if args.fonts:
        [shutil.copy(f, fonts_dir) for f in args.fonts]
        fonts = args.fonts
    elif args.pull_request:
        fonts = download_files_in_github_pr(
            args.pull_request,
            fonts_dir,
            ignore_static_dir=False,
        )
        if not fonts:
            logger.info("No fonts found in pull request. Skipping")
            return
    elif args.github_dir:
        fonts = download_files_in_github_dir(args.github_dir, fonts_dir)
        if not fonts:
            logger.info("No fonts found in github dir. Skipping")
            return
    elif args.archive:
        fonts = download_files_from_archive(args.archive, fonts_dir)
    elif args.googlefonts:
        fonts = download_family_from_Google_Fonts(args.googlefonts, fonts_dir)

    if args.filter_fonts:
        re_filter = re.compile(args.filter_fonts)
        fonts = [f for f in fonts if re_filter.search(f)]

    dfonts = [
        DFont(f) for f in fonts if f.endswith((".ttf", ".otf")) and "static" not in f
    ]
    family_name = family_name_from_fonts(dfonts)
    family_on_gf = Google_Fonts_has_family(family_name)

    # Retrieve fonts_before and store in out dir
    fonts_before = None
    if any(
        [
            args.fonts_before,
            args.pull_request_before,
            args.github_dir_before,
            args.archive_before,
        ]
    ) or (args.googlefonts_before and family_on_gf):
        fonts_before_dir = os.path.join(args.out, "fonts_before")
        mkdir(fonts_before_dir, overwrite=False)
    if args.fonts_before:
        [shutil.copy(f, fonts_before_dir) for f in args.fonts_before]
        fonts_before = args.fonts_before
    elif args.pull_request_before:
        fonts_before = download_files_in_github_pr(
            args.pull_request_before, fonts_before_dir, ignore_static_dir=False
        )
    elif args.github_dir_before:
        fonts_before = download_files_in_github_dir(
            args.github_dir_before, fonts_before_dir
        )
    elif args.archive_before:
        fonts_before = download_files_from_archive(
            args.archive_before, fonts_before_dir
        )
    elif args.googlefonts_before and family_on_gf:
        fonts_before = download_family_from_Google_Fonts(family_name, fonts_before_dir)

    url = None
    if args.out_url:
        url = args.out_url
    elif args.out_github and args.pull_request:
        url = args.pull_request
    elif args.out_github and args.github_dir:
        url = args.github_dir

    if fonts_before:
        dfonts_before = [
            DFont(f)
            for f in fonts_before
            if f.endswith((".ttf", ".otf")) and "static" not in f
        ]
        qa = FontQA(dfonts, dfonts_before, args.out, url=url)
    else:
        qa = FontQA(dfonts, out=args.out, url=url)

    if args.auto_qa and family_on_gf:
        qa.googlefonts_upgrade(args.imgs, args.rust)
    elif args.auto_qa and not family_on_gf:
        qa.googlefonts_new(args.imgs, args.rust)
    if args.render:
        qa.render(args.imgs, args.rust)
    if args.fontbakery:
        if args.rust:
            qa.fontspector(extra_args=args.extra_fontspector_args)
        else:
            qa.fontbakery(extra_args=args.extra_fontbakery_args)
    if args.diffenator:
        if args.rust:
            qa.diffenator3()
        else:
            qa.diffenator()
    if args.diffbrowsers:
        qa.diffbrowsers(args.imgs, rust=args.rust)
    if args.proof:
        qa.proof(rust=args.rust)
    if args.interpolations:
        qa.interpolations(rust=args.rust)

    if qa.has_error:
        logger.fatal("QA tools have raised a fatal error. Please fix!")
        sys.exit(1)


if __name__ == "__main__":
    main()



from gftools.builder.operations import OperationBase


class RemapLayout(OperationBase):
    description = "Run gftools-remap-layout to change a font's layout rules"
    rule = "gftools-remap-layout -o $out $in $args"



#!/usr/bin/env python3
"""
Add and replace tags from another tagging spreadsheet.
"""
import csv
import math
import argparse
from pathlib import Path


def read_csv(filepath):
    """Read CSV file and return list of rows."""
    with open(filepath, "r", encoding="utf-8") as f:
        reader = csv.reader(f)
        return list(reader)


def write_csv(filepath, rows):
    """Write rows to CSV file."""
    with open(filepath, "w", encoding="utf-8", newline="") as f:
        writer = csv.writer(f, lineterminator="\n")
        writer.writerows(rows)


def sort_rows(rows):
    """
    Sort rows by displayName (col 0) then category (col 2), faithful to the JS:
    if (`${a.displayName},${a.category}` < `${b.displayName},${b.category}`)

    The "displayName,category" string is the primary key (kept comma-joined as in
    the JS so family/tag ordering stays byte-identical to existing files). The
    axis column (col 1) is NOT part of the primary key — including it there would
    reshuffle every variable-font row away from the established (family, tag)
    order. Instead, within each (family, tag) group VF rows are ordered by the
    NUMERIC axis value, so e.g. wght@400 precedes wght@800 and wght@900 precedes
    wght@1000.
    """

    data_rows = rows

    def axis_value(axis):
        # Parse "wght@400" -> ("wght", 400.0) so variable-font rows order by the
        # NUMERIC axis value (wght@400 before wght@800, wght@900 before wght@1000).
        # Static rows (empty axis) sort first within a (family, tag) group.
        if "@" in axis:
            name, _, value = axis.partition("@")
            try:
                return (name, float(value))
            except ValueError:
                return (axis, math.inf)
        return ("", float("-inf"))

    def sort_key(row):
        # Primary key "displayName,category" (col 0, col 2 — NOT the axis col 1),
        # kept as the original JS string so family/tag order stays byte-identical.
        # Tiebreak on the axis (col 1) by numeric value within each (family, tag).
        return (f"{row[0]},{row[2]}", axis_value(row[1]))

    sorted_data = sorted(data_rows, key=sort_key)
    return sorted_data


def merge_csvs(source_path, target_path, tag_category, output_path=None):
    """
    Merge source CSV into target CSV.
    - If first 3 columns match, update 4th column only
    - Otherwise, add new row
    - Sort and save result
    """
    source_rows = read_csv(source_path)
    target_rows = read_csv(target_path)

    # Assume first row is header
    if len(source_rows) == 0 or len(target_rows) == 0:
        print("Error: One or both CSV files are empty")
        return

    source_data = {(r[0], r[1], r[2]): r for r in source_rows}
    target_data = {(r[0], r[1], r[2]): r for r in target_rows if r}

    to_add = set(source_data.keys()) - set(target_data.keys())
    to_update = set(source_data.keys()) & set(target_data.keys())

    for key in to_add:
        _, _, category = key
        if category != tag_category:
            continue
        target_data[key] = source_data[key]

    for key in to_update:
        _, _, category = key
        if category != tag_category:
            continue
        target_data[key][3] = source_data[key][3]

    all_rows = list(target_data.values())
    sorted_rows = sort_rows(all_rows)

    output = output_path if output_path else target_path
    write_csv(output, sorted_rows)
    print(f"Merged CSV saved to: {output}")
    print(f"Total rows (including header): {len(sorted_rows)}")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Merge two CSV files with custom logic and sorting"
    )
    parser.add_argument(
        "source", type=Path, help="Source CSV file (rows to merge from)"
    )
    parser.add_argument(
        "target", type=Path, help="Target CSV file (rows to merge into)"
    )
    parser.add_argument("tag_category", help="Tag category to add/replace")
    parser.add_argument(
        "-o", "--output", type=Path, help="Output CSV file (default: overwrite target)"
    )
    args = parser.parse_args(args)

    merge_csvs(args.source, args.target, args.tag_category, args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Rename a font.

Changes font menu name and filename. User can also specify their
own output path.

Usage:
gftools rename-font font.ttf "New Family Name"
"""
import argparse
from fontTools.ttLib import TTFont
from gftools.utils import font_familyname
from gftools.fix import rename_font


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("font")
    parser.add_argument("new_name", help="New family name")
    parser.add_argument("-o", "--out", help="Output path")
    parser.add_argument(
        "--suffix", action="store_true", help="New name is added to old name"
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="New font is written on old filename"
    )
    parser.add_argument(
        "--just-family",
        action="store_true",
        help="Only change family name and names based off it, such as the "
        "PostScript name. (By default, the old family name is replaced "
        "by the new name in all name table entries, including copyright, "
        "description, etc.)",
    )
    args = parser.parse_args(args)

    font = TTFont(args.font)
    current_name = font_familyname(font)
    if args.suffix:
        args.new_name = current_name + args.new_name
    rename_font(font, args.new_name, aggressive=not args.just_family)

    if args.out:
        out = args.out
    elif args.overwrite:
        out = args.font
    else:
        out = args.font.replace(
            current_name.replace(" ", ""), args.new_name.replace(" ", "")
        )
    print("Saving font: {}".format(out))
    font.save(out)


if __name__ == "__main__":
    main()



import logging
import os
from pathlib import Path
import subprocess
import traceback
from typing import List, Sequence

from gftools.gfgithub import GitHubClient
from gftools.utils import mkdir

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def report_exceptions(meth):
    def safe_call(self, *args, **kwargs):
        try:
            meth(self, *args, **kwargs)
        except Exception as e:
            msg = f"Call to {meth.__name__} failed:\n{e}"
            print(msg)
            print()
            print(traceback.format_exc())
            self.post_to_github(msg + "\n\n" + "See CI logs for more details")

    return safe_call


def all_relevant_files(fonts: Sequence[DFont]) -> List[str]:
    """Returns a list of all relevant files for the given fonts."""
    files = []
    relevant_globs = [
        "METADATA.pb",
        "OFL.txt",
        "LICENSE.txt",
        "DESCRIPTION.en_us.html",
        "article/*",
    ]

    for font in fonts:
        path = Path(font.path)
        files.append(str(path))
        for glob in relevant_globs:
            for file in path.parent.glob(glob):
                if str(file) not in files:
                    files.append(str(file))
    return files


class FontQA:
    def __init__(self, fonts, fonts_before=None, out="out", url=None):
        self.fonts = fonts
        self.fonts_before = fonts_before
        self.out = out
        self.url = url
        self.has_error = False

    @report_exceptions
    def diffenator3(self, **kwargs):
        logger.info("Running Diffenator3")
        if not self.fonts_before:
            logger.warning("Cannot run Diffenator since there are no fonts before")
            return
        assert len(self.fonts) >= len(self.fonts_before)
        for f, f_before in zip(
            sorted([f.path for f in self.fonts]),
            sorted([f.path for f in self.fonts_before]),
        ):
            cmd = [
                "diffenator3",
                "--html",
                "--instance",
                "*",
                "--output",
                os.path.join(self.out, "Diffenator"),
                f_before,
                f,
            ]
            process = subprocess.run(cmd)
            if process.returncode != 0:
                self.has_error = True
                return

    @report_exceptions
    def diffbrowsers(self, imgs=False):
        logger.info("Running Diffbrowsers")
        if not self.fonts_before:
            logger.warning("Cannot run diffbrowsers since there are no fonts before")
            return
        dst = os.path.join(self.out, "Diffbrowsers")
        mkdir(dst)
        assert len(self.fonts) >= len(self.fonts_before)
        for f, f_before in zip(
            sorted([f.path for f in self.fonts]),
            sorted([f.path for f in self.fonts_before]),
        ):
            cmd = [
                "diff3proof",
                "--output",
                dst,
                f_before,
                f,
            ]
            process = subprocess.run(cmd)
            if process.returncode != 0:
                self.has_error = True
            os.rename(
                os.path.join(dst, "diff3proof.html"),
                os.path.join(dst, f"diff3proof-{Path(f).stem}.html"),
            )
        return

    @report_exceptions
    def proof(self, imgs=False):
        logger.info("Running proofing tools")
        dst = os.path.join(self.out, "Proof")
        mkdir(dst)
        for font in self.fonts:
            cmd = [
                "diff3proof",
                "--output",
                dst,
                font.path,
            ]
            process = subprocess.run(cmd)
            os.rename(
                os.path.join(dst, "diff3proof.html"),
                os.path.join(dst, f"diff3proof-{Path(font.path).stem}.html"),
            )

            if process.returncode != 0:
                self.has_error = True
                return

    @report_exceptions
    def interpolations(self):
        dst = os.path.join(self.out, "Interpolations")
        if not any(f.is_variable() for f in self.fonts):
            return
        mkdir(dst)
        for font in self.fonts:
            font_dst = os.path.join(dst, f"{os.path.basename(font.path[:-4])}.pdf")
            if not font.is_variable():
                continue
            if rust:
                cmd = ["interpolatable", font.path, "--pdf", font_dst]
            else:
                cmd = [
                    "fonttools",
                    "varLib.interpolatable",
                    font.path,
                    "--pdf",
                    font_dst,
                ]
            subprocess.call(cmd)

    @report_exceptions
    def fontbakery(self, profile="googlefonts", html=False, extra_args=None):
        logger.info("Running Fontbakery")
        out = os.path.join(self.out, "Fontbakery")
        mkdir(out)
        cmd = (
            ["fontbakery", "check-" + profile, "-l", "INFO", "--succinct"]
            + [f.path for f in self.fonts]
            + ["-C"]
            + ["--ghmarkdown", os.path.join(out, "report.md")]
            + ["-e", "FATAL"]
        )
        if html:
            cmd.extend(["--html", os.path.join(out, "report.html")])
        if extra_args:
            cmd.extend(extra_args)
        process = subprocess.run(cmd)

        fontbakery_report = os.path.join(self.out, "Fontbakery", "report.md")
        if not os.path.isfile(fontbakery_report):
            logger.warning(
                "Cannot Post Github message because no Fontbakery report exists"
            )
            return
        with open(fontbakery_report) as doc:
            msg = doc.read()
            self.post_to_github(msg)

        if process.returncode != 0:
            self.has_error = True

    @report_exceptions
    def fontspector(self, profile="googlefonts", html=False, extra_args=None):
        logger.info("Running Fontspector")
        out = os.path.join(self.out, "Fontspector")
        mkdir(out)
        cmd = (
            [
                "fontspector",
                "--profile",
                profile,
                "-l",
                "info",
                "--succinct",
                "-e",
                "error",
            ]
            + all_relevant_files(self.fonts)
            + ["--ghmarkdown", os.path.join(out, "report.md")]
        )
        if html:
            cmd.extend(["--html", os.path.join(out, "report.html")])
        if extra_args:
            cmd.extend(extra_args)
        process = subprocess.run(cmd)

        fontspector_report = os.path.join(self.out, "Fontspector", "report.md")
        if not os.path.isfile(fontspector_report):
            logger.warning(
                "Cannot Post Github message because no Fontspector report exists"
            )
            return
        with open(fontspector_report) as doc:
            msg = doc.read()
            self.post_to_github(msg)

        if process.returncode != 0:
            self.has_error = True

    def googlefonts_upgrade(self, imgs=False):
        self.fontspector()
        self.diffenator3()
        self.diffbrowsers(imgs)
        self.interpolations()

    def googlefonts_new(self, imgs=False):
        self.fontspector()
        self.diffenator3()
        self.proof(imgs)
        self.interpolations()

    def render(self, imgs=False):
        if self.fonts_before:
            self.diffbrowsers(imgs)
        else:
            self.proof(imgs)

    def post_to_github(self, text):
        """Post text as a new issue or as a comment to an open
        PR"""
        if not self.url:
            return
        # Parse url tokens
        url_split = self.url.split("/")
        repo_owner = url_split[3]
        repo_name = url_split[4]
        issue_number = url_split[-1] if "pull" in self.url else None

        client = GitHubClient(repo_owner, repo_name)

        try:
            if issue_number:
                client.create_issue_comment(issue_number, text)
            else:
                client.create_issue("Google Font QA report", text)
        except Exception as e:
            logger.warn(
                "Cannot post QA report!\n"
                "Most likely, the repository may lack a GH_TOKEN secret, or "
                "the pull request has come from a forked repo which "
                "is not allowed to access the repo's secrets for "
                f"security reasons. Full traceback:\n{e}"
            )



from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
import os
import pkgutil
import importlib
import inspect
import platform
import sys
from os.path import dirname
from tempfile import NamedTemporaryFile
from typing import Dict

from gftools.builder.file import File
from gftools.utils import shell_quote


TOUCH = "touch"
if platform.system() == "Windows":
    TOUCH = "echo. >"


@dataclass
class OperationBase:
    postprocess: bool = False
    original: dict = field(default_factory=dict)
    _targets: set = field(default_factory=set)  # [File]
    _sources: set = field(default_factory=set)  # [File]
    implicit: set = field(default_factory=set)

    in_place = False
    description = "A badly described rule"
    rule: str = "echo"  # Must be overridden in subclass

    def __eq__(self, other):
        return self.original == other.original

    # Having a convenience function for "looks equal" is nice but
    # sometimes you actually need to know if it's literally the same
    # step
    def object_equals(self, other):
        return id(self) == id(other)

    def convert_dependencies(self, graph):
        if "needs" in self.original:
            if not isinstance(self.original["needs"], list):
                self.original["needs"] = [self.original["needs"]]
            self.original["needs"] = [
                graph._ensure_named_file(dependency)
                for dependency in self.original["needs"]
            ]

    @classmethod
    def write_rules(cls, writer):
        name = cls.__module__.split(".")[-1]
        writer.comment(name + ": " + cls.description)
        if os.name == "nt":
            cmd = "cmd /c " + cls.rule + " $stamp"
        else:
            cmd = cls.rule + " $stamp"
        writer.rule(
            name,
            f"{shell_quote(sys.executable)} -m gftools.builder.jobrunner {cmd}",
            description=name,
        )
        writer.newline()

    @property
    def opname(self):
        return self.__class__.__module__.split(".")[-1]

    def set_target(self, target):
        self._targets.add(target)

    def set_source(self, source):
        self._sources.add(source)

    @cached_property
    def stamppath(self):
        return NamedTemporaryFile().name + f".{self.opname}stamp"

    @property
    def variables(self):
        return {k: v for k, v in self.original.items() if k != "needs"}

    def build(self, writer):
        if self.postprocess:
            # Check this *is* a post-process step
            stamp = f" && {TOUCH} {self.stamppath}"
            writer.comment(
                "Postprocessing "
                + ", ".join([t.path for t in self.targets])
                + " with "
                + self.__class__.__name__
            )
            writer.build(
                self.stamppath,
                self.opname,
                self.dependencies,
                variables={"stamp": stamp, **self.variables},
                implicit=[
                    t.path for t in self.implicit if t.path not in self.dependencies
                ],
            )
        else:
            writer.comment("Generating " + ", ".join([t.path for t in self.targets]))
            writer.build(
                list(set([t.path for t in self.targets])),
                self.opname,
                self.dependencies,
                variables=self.variables,
                implicit=[
                    t.path for t in self.implicit if t.path not in self.dependencies
                ]
                or None,
            )

    def __hash__(self):
        return hash(id(self))

    @property
    def dependencies(self):
        sources = [source.path for source in self._sources]
        if "needs" in self.original:
            return sources + [d.path for d in self.original["needs"]]
        return list(set(sources))

    @property
    def targets(self):
        return self._targets

    @property
    def first_target(self):
        return list(self.targets)[0]

    @property
    def first_source(self):
        return list(self._sources)[0]

    def validate(self):
        return True


class FontmakeOperationBase(OperationBase):
    @property
    def variables(self):
        vars = defaultdict(str)
        for k, v in self.original.items():
            if k != "needs":
                vars[k] = v

        if self.first_source.is_glyphs:
            vars["fontmake_type"] = "-g"
        elif self.first_source.is_designspace:
            vars["fontmake_type"] = "-m"
        elif self.first_source.is_ufo:
            vars["fontmake_type"] = "-u"
        if "--verbose" not in vars["fontmake_args"]:
            vars["fontmake_args"] += " --verbose WARNING "

        return vars


class OperationRegistry:
    def __init__(self, use_fontc: bool):
        self.known_operations = get_known_operations()
        self.use_fontc = use_fontc

    def get(self, operation_name: str):
        if self.use_fontc:
            if operation_name == "buildVariable":
                # if we import this at the top level it's a circular import error
                from .fontc.fontcBuildVariable import FontcBuildVariable

                return FontcBuildVariable
            if operation_name == "buildTTF":
                from .fontc.fontcBuildTTF import FontcBuildTTF

                return FontcBuildTTF

            if operation_name == "buildOTF":
                from .fontc.fontcBuildOTF import FontcBuildOTF

                return FontcBuildOTF

        return self.known_operations.get(operation_name)


def get_known_operations() -> Dict[str, OperationBase]:
    known_operations = {}

    for mod in pkgutil.iter_modules([dirname(__file__)]):
        if "fontc" in mod.name:
            continue
        imp = importlib.import_module("gftools.builder.operations." + mod.name)
        classes = [
            (name, cls)
            for name, cls in inspect.getmembers(
                sys.modules[imp.__name__], inspect.isclass
            )
            if "OperationBase" not in name and issubclass(cls, OperationBase)
        ]
        if len(classes) > 1:
            raise ValueError(
                f"Too many classes in module gftools.builder.operations.{mod.name}"
            )
        known_operations[mod.name] = classes[0][1]
    return known_operations



#!/usr/bin/env python3
"""
gftools gen-spac

Add a Spacing axis (SPAC) to a variable font.
https://fonts.google.com/variablefonts#axis-definitions

Usage:
gftools gen-spac font.ttf --amount 100 --inplace
"""
from fontTools.ttLib.tables._f_v_a_r import Axis
from fontTools.ttLib.tables.TupleVariation import TupleVariation
from fontTools.ttLib import TTFont
from fontTools.varLib.hvar import add_HVAR
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.ttLib.tables import otTables
import argparse


def add_spacing_axis(font, min_amount, max_amount):
    assert "fvar" in font, "Font must have an 'fvar' table"
    gvar = font["gvar"]
    for glyph_name in font.getGlyphOrder():
        glyph = font["glyf"][glyph_name]
        if not hasattr(glyph, "coordinates"):
            continue
        glyph_variations = gvar.variations.get(glyph_name)
        if not glyph_variations:
            continue
        min_coords = [None] * len(glyph_variations[0].coordinates)
        min_coords[-3] = (min_amount, 0)
        min_coords[-4] = (-min_amount, 0)
        min_tp = TupleVariation({"SPAC": (-1.0, -1.0, 0.0)}, min_coords)
        gvar.variations[glyph_name].append(min_tp)

        max_coords = [None] * len(glyph_variations[0].coordinates)
        max_coords[-3] = (max_amount, 0)
        max_coords[-4] = (-max_amount, 0)
        max_tp = TupleVariation({"SPAC": (0.0, 1.0, 1.0)}, max_coords)
        gvar.variations[glyph_name].append(max_tp)

    name_table = font["name"]

    axis = Axis()
    axis.axisTag = "SPAC"
    axis.axisNameID = name_table.addMultilingualName({"en": "Spacing"})
    axis.minValue = min_amount
    axis.defaultValue = 0.0
    axis.maxValue = max_amount

    fvar = font["fvar"]
    fvar.axes.append(axis)
    for inst in fvar.instances:
        inst.coordinates["SPAC"] = 0.0
    add_HVAR(font)

    print("Update various VarStores")
    stores = []
    for table in ("MVAR", "HVAR", "BASE", "VVAR", "COLR", "GDEF"):
        if table in font and hasattr(font[table].table, "VarStore"):
            stores.append(font[table].table.VarStore)
    spacRegion = otTables.VarRegionAxis()
    spacRegion.StartCoord = -1
    spacRegion.PeakCoord = 0
    spacRegion.EndCoord = 1
    for store in stores:
        store.VarRegionList.RegionAxisCount = len(fvar.axes)
        for region in store.VarRegionList.Region:
            while len(region.VarRegionAxis) < len(fvar.axes):
                region.VarRegionAxis.append(spacRegion)


def main(args=None):
    parser = argparse.ArgumentParser(description="Add a SPAC axis to a variable font.")
    parser.add_argument(
        "font", type=TTFont, help="Path to the variable font file to modify."
    )
    parser.add_argument(
        "min",
        type=int,
        help="Min amount of spacing to add",
    )
    parser.add_argument(
        "max",
        type=int,
        help="Max amount of spacing to add",
    )
    out_group = parser.add_mutually_exclusive_group(required=False)
    out_group.add_argument("--out", "-o", help="Output dir for fonts")
    out_group.add_argument(
        "--inplace", action="store_true", default=False, help="Overwrite input files"
    )
    args = parser.parse_args(args)

    add_spacing_axis(args.font, args.min, args.max)
    if args.inplace:
        output_path = args.font.reader.file.name
    elif args.out:
        output_path = args.out
    else:
        output_path = makeOutputFileName(args.font.reader.file.name)
    args.font.save(output_path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility to print unicode character names from a nam file.

Input file should have one codepoint per line in hex (0xXXXX).

"""
from __future__ import print_function
import unicodedata
import sys
import argparse

parser = argparse.ArgumentParser(
    description="Add Unicode character names to a nam file"
)
parser.add_argument("--nam_file", help="Location of nam file")


def main(args=None):
    args = parser.parse_args(args)
    with open(args.nam_file, "r") as f:
        for line in f:
            print(_ReformatLine(line))


def _ReformatLine(line):
    if line.startswith("0x"):
        codepoint = int(line[2:6], 16)  # This'll only work for BMP...
        out = chr(codepoint) + " " + unicodedata.name(chr(codepoint), "")
        return "0x%04X  %s" % (codepoint, out)
    else:
        return line


if __name__ == "__main__":
    main()



from tempfile import TemporaryDirectory
import copy
import logging
import os
import re
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple

from glyphsLib.builder import UFOBuilder
import yaml
from fontTools.designspaceLib import InstanceDescriptor
from strictyaml import load, YAMLValidationError

from gftools.builder.file import File
from gftools.builder.recipeproviders import RecipeProviderBase
from gftools.builder.schema import (
    GOOGLEFONTS_SCHEMA,
    stat_schema,
    avar1_schema,
    avar2_schema,
    stat_schema_by_font_name,
)
from gftools.utils import open_ufo

logger = logging.getLogger("GFBuilder")

Italic = Optional[Tuple[str, float, float]]  # tag, regular value, italic value

# Things not ported from old builder:
#  - cleanup
#  - stylespace (old, replaced by stat)
#  - axisOrder (did not actually do anything)
#  - googleFonts (only used by actions, not builder)
#  - interpolate (always on for statics now)
#  - useMutatorMath (dead in fontmake)

# Taken from gftools-builder
DEFAULTS = {
    "outputDir": "../fonts",
    "vfDir": "$outputDir/variable",
    "ttDir": "$outputDir/ttf",
    "otDir": "$outputDir/otf",
    "woffDir": "$outputDir/webfonts",
    "buildStatic": True,
    "buildOTF": True,
    "buildTTF": True,
    "buildSmallCap": True,
    "autohintTTF": True,
    "autohintOTF": False,
    "ttfaUseScript": False,
    "logLevel": "WARN",
    "cleanUp": True,
    "includeSourceFixes": False,
    "fvarInstanceAxisDflts": None,
    "flattenComponents": True,
    "addGftoolsVersion": True,
    "decomposeTransformedComponents": True,
    "interpolate": False,
    "useMutatorMath": False,
    "checkCompatibility": True,
    "overlaps": "booleanOperations",
    "splitItalic": True,
}


class GFBuilder(RecipeProviderBase):
    schema = GOOGLEFONTS_SCHEMA

    def revalidate(self):
        # Revalidate using our schema
        try:
            load(self.builder._orig_config, self.schema)
        except YAMLValidationError as e:
            raise ValueError("Invalid configuration file") from e

    def write_recipe(self):
        if "instances" in self.config:
            logger.warning(
                "'instances' no longer supported; generate a config with --generate and select the instances you want"
            )
        self.revalidate()
        self.config = {**DEFAULTS, **self.config}
        for field in ["vfDir", "ttDir", "otDir", "woffDir"]:
            self.config[field] = self.config[field].replace(
                "$outputDir", self.config["outputDir"]
            )
        self.config["buildWebfont"] = self.config.get(
            "buildWebfont", self.config.get("buildStatic", True)
        )

        if "stat" in self.config:
            self.statfile = NamedTemporaryFile(delete=False, mode="w+")
            try:
                load(yaml.dump(self.config["stat"]), stat_schema)
            except YAMLValidationError:
                load(yaml.dump(self.config["stat"]), stat_schema_by_font_name)
                if self.config.get("buildSmallCap"):
                    for font in list(self.config["stat"].keys()):
                        scfont = re.sub(r"((?:-Italic)?\[)", r"SC\1", font)
                        self.config["stat"][scfont] = self.config["stat"][font]
            yaml.dump(self.config["stat"], self.statfile)
        else:
            self.statfile = None

        if "avar2" in self.config:
            self.avar2file = NamedTemporaryFile(delete=False, mode="w+")
            load(yaml.dump(self.config["avar2"]), avar2_schema)
            for font in list(self.config["avar2"].keys()):
                scfont = re.sub(r"((?:-Italic)?\[)", r"SC\1", font)
                self.config["avar2"][scfont] = self.config["avar2"][font]
            yaml.dump(self.config["avar2"], self.avar2file)
        else:
            self.avar2file = None

        if "avar1" in self.config:
            load(yaml.dump(self.config["avar1"]), avar1_schema)

        if "fvarInstances" in self.config:
            self.fvarInstancesFile = NamedTemporaryFile(delete=False, mode="w+")
            for font in list(self.config["fvarInstances"].keys()):
                scfont = re.sub(r"((?:-Italic)?\[)", r"SC\1", font)
                self.config["fvarInstances"][scfont] = self.config["fvarInstances"][
                    font
                ]
            yaml.dump(self.config["fvarInstances"], self.fvarInstancesFile)
        else:
            self.fvarInstancesFile = None

        # Find variable fonts
        self.recipe = {}
        self.build_all_variables()
        self.build_all_statics()
        return self.recipe

    def _has_slant_ital(self, source: File) -> Italic:
        if source.is_glyphs:
            tags = [ax.axisTag for ax in source.gsfont.axes]
        elif source.is_designspace:
            tags = [ax.tag for ax in source.designspace.axes]
        else:
            return
        if "ital" in tags:
            slanty_axis = "ital"
        elif "slnt" in tags:
            slanty_axis = "slnt"
        else:
            return
        if source.is_glyphs:
            gsfont = source.gsfont
            builder = UFOBuilder(gsfont, minimal=True)
            builder.to_designspace_axes()
            axes = builder.designspace.axes
        else:
            axes = source.designspace.axes
        wanted = [axis for axis in axes if axis.tag == slanty_axis]
        if slanty_axis == "ital":
            return (slanty_axis, wanted[0].minimum, wanted[0].maximum)
        else:
            # We expect the italic value to have negative slant, so it
            # turns out as the minimum.
            return (slanty_axis, wanted[0].maximum, wanted[0].minimum)

    def _vf_filename(
        self, source, suffix="", extension="ttf", italic_ds=None, roman=False
    ):
        """Determine the file name for a variable font."""
        sourcebase = os.path.splitext(source.basename)[0]
        if source.is_glyphs_file:
            # Optimisation: avoid parsing the full GSFont here
            if source.glyphs_format >= 3:
                tags = [ax["tag"] for ax in source.glyphs_plist["axes"]]
            else:
                axes = next(
                    param["value"]
                    for param in source.glyphs_plist["customParameters"]
                    if param["name"] == "Axes"
                )
                tags = [axis["Tag"] for axis in axes]
        elif source.is_glyphspackage:
            # Optimisation: avoid parsing the full GSFont here when what we need
            # is within the fontinfo.plist
            tags = [ax["tag"] for ax in source.glyphspackage_fontinfo["axes"]]
        elif source.is_designspace:
            tags = [ax.tag for ax in source.designspace.axes]
        else:
            raise ValueError("Unknown source type")

        if "spacingAxis" in self.config:
            tags.append("SPAC")

        if italic_ds:
            if not roman:
                sourcebase += "-Italic"
            tags.remove(italic_ds[0])
        axis_tags = ",".join(sorted(tags))
        directory = self.config["vfDir"]
        if extension == "woff2":
            directory = self.config["woffDir"]
        if suffix:
            # Put any suffix before the -Italic element
            if "-Italic" in sourcebase:
                sourcebase = sourcebase.replace("-Italic", suffix + "-Italic")
            else:
                sourcebase += suffix

        return os.path.join(directory, f"{sourcebase}[{axis_tags}].{extension}")

    def _static_filename(
        self, instance: InstanceDescriptor, suffix: str = "", extension: str = "ttf"
    ):
        """Determine the file name for a static font."""
        if extension == "ttf":
            outdir = self.config["ttDir"]
        elif extension == "otf":
            outdir = self.config["otDir"]
        elif extension == "woff2":
            outdir = self.config["woffDir"]
        instancebase = os.path.splitext(os.path.basename(instance.filename))[0]
        if suffix:
            # This is horrible; insert the suffix at the end of the family
            # name, before the style name.
            familyname, path = instancebase.rsplit("-", 1)
            instancebase = familyname + suffix + "-" + path

        return os.path.join(outdir, f"{instancebase}.{extension}")

    def fontmake_args(self, source, variable=False):
        args = "--filter ... "
        if self.config.get("flattenComponents", True):
            args += " --filter FlattenComponentsFilter"
        if self.config.get("decomposeTransformedComponents", True):
            args += " --filter DecomposeTransformedComponentsFilter"

        if self.config.get("logLevel") != "WARN":
            args += " --verbose " + self.config["logLevel"]
        if self.config.get("reverseOutlineDirection", True) is False:
            args += " --keep-direction"
        if self.config.get("removeOutlineOverlaps") is False:
            args += " --keep-overlaps"
        if self.config.get("expandFeaturesToInstances"):
            args += " --expand-features-to-instances"
        if self.config.get("extraFontmakeArgs") is not None:
            args += " " + str(self.config["extraFontmakeArgs"])
        if variable:
            if self.config.get("checkCompatibility") is False:
                args += " --no-check-compatibility"
            if self.config.get("extraVariableFontmakeArgs") is not None:
                args += " " + str(self.config["extraVariableFontmakeArgs"])
            if source.is_glyphs:
                for gd in self.config.get("glyphData", []):
                    args += " --glyph-data " + gd
        else:
            if self.config.get("extraStaticFontmakeArgs") is not None:
                args += " " + str(self.config["extraStaticFontmakeArgs"])
        return args

    def fix_args(self):
        args = ""
        if self.config.get("includeSourceFixes"):
            args += " --include-source-fixes"
        if self.config.get("fvarInstanceAxisDflts"):
            args += (
                " --fvar-instance-axis-dflts '"
                + " ".join(
                    [
                        f"{k}={v}"
                        for k, v in self.config["fvarInstanceAxisDflts"].items()
                    ]
                )
                + "'"
            )
        return args

    def build_all_variables(self):
        if not self.config.get("buildVariable", True):
            return
        for source in self.sources:
            if not source.is_variable:
                continue
            italic_ds = None
            if self.config["splitItalic"]:
                italic_ds = self._has_slant_ital(source)
            if italic_ds:
                self.build_a_variable(source, italic_ds=italic_ds, roman=True)
                self.build_a_variable(source, italic_ds=italic_ds, roman=False)
                if self.statfile:
                    self._italicize_stat_file(source, italic_ds)
            else:
                self.build_a_variable(source)
        self.build_STAT()
        if "spacingAxis" in self.config:
            self.build_spacing_axis()
        if "avar2" in self.config:
            self.build_avar2()
        if "fvarInstances" in self.config:
            self.build_fvar_instances()
        if "avar1" in self.config:
            self.build_avar1()

    def build_spacing_axis(self):
        vfs = [x for x in self.recipe.keys() if x.endswith("ttf")]
        if len(vfs) > 0:
            args = {
                "args": str(self.config["spacingAxis"]["min"])
                + " "
                + str(self.config["spacingAxis"]["max"]),
                "postprocess": "addSpacingAxis",
            }
            for vf in vfs:
                self.recipe[vf].append(args)

    def build_STAT(self):
        # Add buildStat to a variable target, it'll do for all of them

        # We've built a bunch of variables here, but we may also have
        # some woff2 we added as part of the process, so ignore them.
        all_variables = [x for x in self.recipe.keys() if x.endswith("ttf")]
        if len(all_variables) > 0:
            last_target = all_variables[-1]
            if self.statfile:
                self.statfile.close()
                args = {"args": "--src " + self.statfile.name}
            else:
                args = {}
            other_variables = list(set(all_variables) - set([last_target]))
            build_stat_step = {
                "postprocess": "buildStat",
                **args,
            }
            if other_variables:
                build_stat_step["needs"] = other_variables
            self.recipe[last_target].append(build_stat_step)

    def build_avar2(self):
        vfs = [x for x in self.recipe.keys() if x.endswith("ttf")]
        if len(vfs) > 0:
            args = {"args": self.avar2file.name, "postprocess": "buildAvar2"}
            for vf in vfs:
                self.recipe[vf].append(args)
            self.avar2file.close()

    def build_avar1(self):
        # Flatten avar2 variable fonts into avar1-only variable fonts,
        # placed in an avar1/ subdirectory next to the avar2 original.
        for fontfile, all_settings in self.config["avar1"].items():
            vf = next((t for t in self.recipe if os.path.basename(t) == fontfile), None)
            if vf is None:
                raise ValueError(f"avar1: no variable font target named '{fontfile}'")
            if not isinstance(all_settings, list):
                all_settings = [all_settings]
            for settings in all_settings:
                self._build_an_avar1(vf, fontfile, settings or {})

    def _build_an_avar1(self, vf, fontfile, settings):
        if "tolerance" in settings:
            args = ["--tolerance", str(settings["tolerance"])]
        else:
            args = ["--no-verify"]
        newname = fontfile
        if "axes" in settings:
            axes = settings["axes"].replace(" ", "")
            args += ["--axes", axes]
            # Redo the [axes] part of the filename for the kept axes
            newname = re.sub(
                r"\[[^]]*\]",
                "[%s]" % ",".join(sorted(axes.split(","))),
                fontfile,
            )
        if "grid" in settings:
            args += ["--grid", settings["grid"].replace(" ", "")]
        if "gridCuts" in settings:
            args += ["--grid-cuts", str(settings["gridCuts"])]
        if "maxMasters" in settings:
            args += ["--max-masters", str(settings["maxMasters"])]
        if "args" in settings:
            args.append(settings["args"])
        target = os.path.join(os.path.dirname(vf), "avar1", newname)
        self.recipe[target] = [
            {"source": vf},
            {"operation": "avar2ToAvar1", "args": " ".join(args)},
        ]

    def build_fvar_instances(self):
        vfs = [x for x in self.recipe.keys() if x.endswith("ttf")]
        if len(vfs) > 0:
            args = {
                "args": self.fvarInstancesFile.name,
                "postprocess": "buildFvarInstances",
            }
            for vf in vfs:
                self.recipe[vf].append(args)
            self.fvarInstancesFile.close()

    def _vtt_steps(self, target: str):
        if os.path.basename(target) in self.config.get("vttSources", {}):
            return [
                {
                    "operation": "buildVTT",
                    "vttfile": self.config["vttSources"][os.path.basename(target)],
                }
            ]
        return []

    def build_a_variable(
        self, source: File, italic_ds: Italic = None, roman: bool = False
    ):
        suffix = self.config.get("filenameSuffix", "")
        if roman:
            target = self._vf_filename(
                source, suffix=suffix, italic_ds=italic_ds, roman=True
            )
        else:
            target = self._vf_filename(
                source, suffix=suffix, italic_ds=italic_ds, roman=False
            )
        steps = (
            [{"source": source.path}]
            + self._subset_steps(source, None)
            + [
                {
                    "operation": "buildVariable",
                    "args": self.fontmake_args(source, variable=True),
                },
            ]
            + self.config.get("postCompile", [])
            + self._vtt_steps(target)
        )
        if italic_ds:
            desired_slice = italic_ds[0] + "="
            if roman:
                desired_slice += str(italic_ds[1])
                steps += [{"operation": "subspace", "axes": desired_slice}]
            else:
                desired_slice += str(italic_ds[2])
                steps += [
                    {"operation": "subspace", "axes": desired_slice},
                ] + self._italic_fixup()

        steps += self._fix_step()

        self.recipe[target] = steps
        self.build_a_webfont(target, self._vf_filename(source, extension="woff2"))
        if self._do_smallcap(source):
            self.recipe[
                self._vf_filename(source, italic_ds=italic_ds, roman=roman, suffix="SC")
            ] = self._smallcap_steps(source, target)

    def _subset_steps(
        self, source: File, instance: Optional[InstanceDescriptor] = None
    ):
        if not self.config.get("includeSubsets"):
            return []
        if not hasattr(self, "_subset_tempdir"):
            self._subset_tempdir = TemporaryDirectory()  # Stow object
        steps = []
        if source.is_glyphs:
            steps.append(
                {
                    "operation": "glyphs2ds",  # GlyphData?
                }
            )
        steps += [
            {
                "operation": "addSubset",
                "subsets": self.config["includeSubsets"],
                "directory": self._subset_tempdir.name,
                "args": "--allow-sparse",
            }
        ]
        if instance:
            steps += [
                {
                    "operation": "instantiateUfo",
                    "instance_name": instance.name,
                    "glyphData": self.config.get("glyphData"),
                    "target": os.path.join(
                        self._subset_tempdir.name,
                        instance.filename + ".json",
                    ),
                }
            ]
        return steps

    def build_all_statics(self):
        if not self.config.get("buildStatic", True):
            return
        for source in self.sources:
            for instance in source.instances:
                if self.config["buildTTF"]:
                    self.build_a_static(source, instance, output="ttf")
                if self.config["buildOTF"]:
                    self.build_a_static(source, instance, output="otf")

    def build_a_static(self, source: File, instance: InstanceDescriptor, output):
        suffix = self.config.get("filenameSuffix", "")
        target = self._static_filename(instance, suffix=suffix, extension=output)

        steps = [
            {"source": source.path},
        ] + self._subset_steps(
            source,
            instance,
        )
        # We skip conversion to UFO if we're running fontc
        use_fontc = self.config.get("use_fontc", False)
        if (
            not source.is_ufo
            and not self.config.get("includeSubsets")
            and not use_fontc
        ):
            instancename = instance.name
            if instancename is None:
                if not instance.familyName or not instance.styleName:
                    raise ValueError(
                        f"Instance {instance.filename} must have a name, or familyName and styleName"
                    )
                instancename = instance.familyName + " " + instance.styleName
            steps.append(
                {
                    "operation": "instantiateUfo",
                    "instance_name": instancename,
                    "glyphData": self.config.get("glyphData"),
                }
            )
        steps += (
            [
                {
                    "operation": "buildTTF" if output == "ttf" else "buildOTF",
                    "args": self.fontmake_args(source, variable=False),
                }
            ]
            + self.config.get("postCompile", [])
            + self._autohint_steps(target)
            + self._vtt_steps(target)
            + self._fix_step()
        )
        self.recipe[target] = steps
        self.build_a_webfont(target, self._static_filename(instance, extension="woff2"))
        if self._do_smallcap(source):
            self.recipe[
                self._static_filename(instance, extension=output, suffix="SC")
            ] = self._smallcap_steps(source, target)

    def build_a_webfont(self, original_target, wf_filename):
        if not self.config["buildWebfont"]:
            return
        if not original_target.endswith(".ttf"):
            return
        self.recipe[wf_filename] = copy.deepcopy(self.recipe[original_target]) + [
            {"operation": "compress"}
        ]

    def _autohint_steps(self, target):
        if bool(self.config.get("autohintTTF")) and target.endswith("ttf"):
            args = "--fail-ok "
            if self.config.get("ttfaUseScript"):
                args += " --auto-script"
            return [{"operation": "autohint", "args": args}]
        if bool(self.config.get("autohintOTF")) and target.endswith("otf"):
            return [{"operation": "autohintOTF"}]
        return []

    def _fix_step(self):
        return [{"operation": "fix", "args": self.fix_args()}]

    def _do_smallcap(self, source):
        if not self.config.get("buildSmallCap"):
            return False
        if source.is_glyphs:
            return "smcp" in source.gsfont.features
        elif source.is_designspace:
            source.designspace.loadSourceFonts(open_ufo)
            return "feature smcp" in source.designspace.sources[0].font.features.text
        else:  # UFO
            return "feature smcp" in open_ufo(source.path).features.text

    def _smallcap_steps(self, source, original):
        new_family_name = source.family_name + " SC"
        return [
            {"source": original},
            {"operation": "remapLayout", "args": "'smcp -> ccmp'"},
            {
                "operation": "rename",
                "args": "--just-family",
                "name": new_family_name,
            },
            {"operation": "fix", "args": self.fix_args()},
        ]

    def _italicize_stat_file(self, source: File, italic_ds: Italic):
        # In this situation we have a stat file, and we have a font with
        # either an ital or a slnt axis that we have split into two subspaced
        # VFs. We now need to rewrite the stat file to remove the slnt axis,
        # and potentially to copy the STAT table to the newly created file.

        # What kind of STAT file are we? A global one or a font-specific one?
        if isinstance(self.config["stat"], dict):
            old_font = self._vf_filename(source)
            new_font = self._vf_filename(source, italic_ds=italic_ds, roman=False)
            if old_font in self.config["stat"] and new_font not in self.config["stat"]:
                raise ValueError(
                    f"We are splitting the font on the {italic_ds[0]} axis, "
                    "but the stat: entry in the config file does not contain "
                    f"an entry for {new_font}. Please add one and try again."
                )
            # Presume the user has done the right thing
            return
        # This is easy, just drop slnt
        self.config["stat"] = [
            axis for axis in self.config["stat"] if axis["tag"] != "slnt"
        ]
        # Rewrite the stat file
        self.statfile.seek(0)
        self.statfile.truncate(0)
        yaml.dump(self.config["stat"], self.statfile)

    def _italic_fixup(self):
        # We have a font created by subspacing the ital or slnt axis, but its
        # name table is not italic yet (and we can't use --update-name-table
        # because we don't have a STAT table eyt). So we need to make this font
        # "italic enough" to convince gftools-fix-font to apply all its italic
        # font fixes (post.italicAngle etc.) when we call it with
        # --include-source-fixes.
        configfile = NamedTemporaryFile(delete=False, mode="w+")
        family_name = self.sources[0].family_name.replace(" ", "")
        # Since this is mad YAML, we can't use the normal YAML library
        # to write this. We'll just write it out manually.
        configfile.write(
            f"""
OS/2->fsSelection: 129
head->macStyle: "|= 0x02"
name->setName: ["{family_name}Italic", 25, 3, 1, 0x409]
name->setName: ["Italic", 2, 3, 1, 0x409]
name->setName: ["Italic", 17, 3, 1, 0x409]
        """
        )
        configfile.close()
        return [
            {
                "operation": "exec",
                "exe": "gftools-fontsetter",
                "args": "-o $out $in " + configfile.name,
            },
            {
                "operation": "fix",
                "args": "--include-source-fixes",
            },
        ]



"""
Functions to fix fonts so they conform to the Google Fonts
specification:
https://github.com/googlefonts/gf-docs/tree/main/Spec
"""

import itertools
from typing import List, Sequence, Tuple
from fontTools.misc.fixedTools import otRound
from fontTools import ttLib
from fontTools.ttLib import TTFont, newTable, getTableModule
from fontTools.ttLib.tables import ttProgram
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
from fontTools.ttLib.tables._f_v_a_r import NamedInstance
from gftools.util.google_fonts import _KNOWN_WEIGHTS
from gftools.utils import (
    download_family_from_Google_Fonts,
    Google_Fonts_has_family,
    font_stylename,
    font_familyname,
    family_bounding_box,
    get_unencoded_glyphs,
    normalize_unicode_marks,
    partition_cmap,
    typo_metrics_enabled,
    validate_family,
    unique_name,
)
from axisregistry import (
    build_filename,
    build_name_table,
    build_fvar_instances,
    build_variations_ps_name,
)
from gftools.stat import gen_stat_tables

from os.path import basename, splitext
from copy import deepcopy
import logging
import math
import subprocess
import os
import tempfile
from datetime import datetime
import re
from gftools.constants import OFL_BODY_TEXT


log = logging.getLogger(__name__)


__all__ = [
    "remove_tables",
    "add_dummy_dsig",
    "fix_unhinted_font",
    "fix_hinted_font",
    "fix_fs_type",
    "fix_weight_class",
    "fix_fs_selection",
    "fix_mac_style",
    "fix_fvar_instances",
    "fix_nametable",
    "inherit_vertical_metrics",
    "fix_vertical_metrics",
    "fix_ascii_fontmetadata",
    "drop_nonpid0_cmap",
    "drop_mac_cmap",
    "fix_pua",
    "fix_isFixedPitch",
    "drop_mac_names",
    "drop_superfluous_mac_names",
    "fix_font",
    "fix_family",
    "rename_font",
    "fix_filename",
]


# The _KNOWN_WEIGHT_VALUES constant is used internally by the GF Engineering
# team so we cannot update ourselves. TODO (Marc F) unify this one day
WEIGHT_NAMES = _KNOWN_WEIGHTS
del WEIGHT_NAMES[""]
WEIGHT_NAMES["Hairline"] = 1
WEIGHT_NAMES["ExtraBlack"] = 1000
WEIGHT_VALUES = {v: k for k, v in WEIGHT_NAMES.items()}

FixResult = Tuple[bool, List[str]]

UNWANTED_TABLES = frozenset(
    [
        "FFTM",
        "TTFA",
        "TSI0",
        "TSI1",
        "TSI2",
        "TSI3",
        "TSI5",
        "prop",
        "Debg",
    ]
)


# Wrapper to communicate metadata
def fixes(*fb_ids):
    def _predicate(func):
        func.fb_ids = fb_ids
        return func

    return _predicate


def _expect(
    ttFont: TTFont, table: str, field: str, value: any, getter=None, setter=None
) -> FixResult:
    """Check that a table has a field with the expected value"""
    if getter is not None:
        previous = getter(ttFont)
    else:
        previous = getattr(ttFont[table], field)
    if previous == value:
        return False, []
    if setter is not None:
        setter(ttFont, value)
    else:
        setattr(ttFont[table], field, value)
    return True, [f"Set {table}.{field} to {value} (was {previous})"]


def _combine_results(*results: FixResult) -> FixResult:
    """Combine multiple FixResults into a single FixResult"""
    return any(r[0] for r in results), list(itertools.chain(*[r[1] for r in results]))


@fixes("com.google.fonts/check/unwanted_tables")
def remove_tables(ttFont: TTFont, tables=None) -> FixResult:
    """Remove unwanted tables from a font. The unwanted tables must belong
    to the UNWANTED_TABLES set.

    Args:
        ttFont: a TTFont instance
        tables: an iterable containing tables remove
    """
    tables_to_remove = UNWANTED_TABLES if not tables else frozenset(tables)
    font_tables = frozenset(ttFont.keys())

    tables_not_in_font = tables_to_remove - font_tables
    if tables_not_in_font:
        log.warning(
            f"Cannot remove tables '{list(tables_not_in_font)}' since they are "
            f"not in the font."
        )

    required_tables = tables_to_remove - UNWANTED_TABLES
    if required_tables:
        log.warning(
            f"Cannot remove tables '{list(required_tables)}' since they are required"
        )

    tables_to_remove = UNWANTED_TABLES & font_tables & tables_to_remove
    if not tables_to_remove:
        return False, []
    for tbl in tables_to_remove:
        del ttFont[tbl]
    return True, [f"Removed tables '{list(tables_to_remove)}'"]


def add_dummy_dsig(ttFont: TTFont) -> FixResult:
    """Add a dummy dsig table to a font. Older versions of MS Word
    require this table.

    Args:
        ttFont: a TTFont instance
    """
    log.warning(
        "Adding dummy DSIG table - the current recommendation is *not* to do this"
    )
    newDSIG = newTable("DSIG")
    newDSIG.ulVersion = 1
    newDSIG.usFlag = 0
    newDSIG.usNumSigs = 0
    newDSIG.signatureRecords = []
    ttFont.tables["DSIG"] = newDSIG
    return ttFont, ["Added dummy DSIG table"]


@fixes("com.google.fonts/check/gasp")
def fix_unhinted_font(ttFont: TTFont) -> FixResult:
    """Improve the appearance of an unhinted font on Win platforms by:
        - Add a new GASP table with a newtable that has a single
          range which is set to smooth.
        - Add a new prep table which is optimized for unhinted fonts.

    Args:
        ttFont: a TTFont instance
    """
    if "fpgm" in ttFont or ("gasp" in ttFont and "prep" in ttFont):
        return False, []
    gasp = newTable("gasp")
    # Set GASP so all sizes are smooth
    gasp.gaspRange = {0xFFFF: 15}

    program = ttProgram.Program()
    assembly = ["PUSHW[]", "511", "SCANCTRL[]", "PUSHB[]", "4", "SCANTYPE[]"]
    program.fromAssembly(assembly)

    prep = newTable("prep")
    prep.program = program

    ttFont["gasp"] = gasp
    ttFont["prep"] = prep
    return True, ["Added GASP and prep tables for unhinted font"]


@fixes("com.google.fonts/check/integer_ppem_if_hinted")
def fix_hinted_font(ttFont: TTFont) -> FixResult:
    """Improve the appearance of a hinted font on Win platforms by enabling
    the head table's flag 3.

    Args:
        ttFont: a TTFont instance
    """
    if not "fpgm" in ttFont:
        return False, []
    return _expect(ttFont, "head", "flags", ttFont["head"].flags | 1 << 3)


@fixes("com.google.fonts/check/fstype")
def fix_fs_type(ttFont: TTFont) -> FixResult:
    """Set the OS/2 table's fsType flag to 0 (Installable embedding).

    Args:
        ttFont: a TTFont instance
    """
    return _expect(ttFont, "OS/2", "fsType", 0)


@fixes("com.google.fonts/check/usweightclass")
def fix_weight_class(ttFont: TTFont) -> FixResult:
    """Set the OS/2 table's usWeightClass so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" in ttFont:
        fvar = ttFont["fvar"]
        default_axis_values = {a.axisTag: a.defaultValue for a in fvar.axes}
        v = default_axis_values.get("wght", None)

        if v is not None:
            return _expect(ttFont, "OS/2", "usWeightClass", int(v))

    stylename = font_stylename(ttFont)
    tokens = stylename.split()
    # Order WEIGHT_NAMES so longest names are first
    for style in sorted(WEIGHT_NAMES, key=len, reverse=True):
        if style in tokens:
            return _expect(ttFont, "OS/2", "usWeightClass", WEIGHT_NAMES[style])

    if "Italic" in tokens:
        return _expect(ttFont, "OS/2", "usWeightClass", 400)
    raise ValueError(
        f"Cannot determine usWeightClass because font style, '{stylename}' "
        f"doesn't have a weight token which is in our known "
        f"weights, '{WEIGHT_NAMES.keys()}'"
    )


@fixes(
    "com.google.fonts/check/os2/use_typo_metrics", "com.google.fonts/check/fsselection"
)
def fix_fs_selection(ttFont: TTFont) -> FixResult:
    """Fix the OS/2 table's fsSelection so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    fs_selection = ttFont["OS/2"].fsSelection

    # turn off all bits except for bit 7 (USE_TYPO_METRICS)
    fs_selection &= 1 << 7

    if "Italic" in tokens:
        fs_selection |= 1 << 0
    if "Bold" in tokens:
        fs_selection |= 1 << 5
    # enable Regular bit for all other styles
    if not tokens & set(["Bold", "Italic"]):
        fs_selection |= 1 << 6
    return _expect(ttFont, "OS/2", "fsSelection", fs_selection)


def fix_mac_style(ttFont: TTFont) -> FixResult:
    """Fix the head table's macStyle so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    mac_style = 0
    if "Italic" in tokens:
        mac_style |= 1 << 1
    if "Bold" in tokens:
        mac_style |= 1 << 0
    return _expect(ttFont, "head", "macStyle", mac_style)


@fixes("com.google.fonts/check/fvar_instances")
def fix_fvar_instances(ttFont, axis_dflts=None) -> FixResult:
    """Replace a variable font's fvar instances with a set of new instances
    that conform to the Google Fonts instance spec:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#fvar-instances

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" not in ttFont:
        return False, []

    if "wght" not in [a.axisTag for a in ttFont["fvar"].axes]:
        return False, []

    fvar = ttFont["fvar"]
    old_instances = {
        (
            ttFont["name"].getDebugName(inst.subfamilyNameID) or "<removed>"
        ): inst.coordinates
        for inst in fvar.instances
    }
    build_fvar_instances(ttFont, axis_dflts)
    new_instances = {
        ttFont["name"].getDebugName(inst.subfamilyNameID): inst.coordinates
        for inst in fvar.instances
    }
    if old_instances != new_instances:
        return True, [
            "Set instances in fvar table to: %s" % ", ".join(new_instances.keys()),
            "(Old instances were: %s)" % ", ".join(old_instances.keys()),
            "Consider fixing export list in source\n",
        ]
    return False, []


@fixes("com.google.fonts/check/font_names")
def fix_nametable(ttFont) -> FixResult:
    """Fix a static font's name table so it conforms to the Google Fonts
    supported styles table:
    https://github.com/googlefonts/gf-docs/tree/main/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    old_nametable = {n.nameID: n.toStr() for n in ttFont["name"].names}
    build_name_table(ttFont)
    new_nametable = {n.nameID: n.toStr() for n in ttFont["name"].names}

    if old_nametable == new_nametable:
        return False, []

    messages = ["Name table entries changed (consider fixing the source instead):"]
    for nid, old_name in old_nametable.items():
        new_name = new_nametable.get(nid, "<removed>")
        if new_name != old_name:
            messages.append("- %i: %s" % (nid, old_name))
            messages.append("+ %i: %s" % (nid, new_name))
    return True, messages


def rename_font(font, new_name, aggressive=True):
    current_name = font_familyname(font)
    if not current_name:
        raise Exception(
            "Name table does not contain nameID 1 or nameID 16. "
            "This tool does not work on webfonts."
        )
    log.info("Updating font name records")
    build_name_table(font, family_name=new_name, aggressive=aggressive)


def fix_filename(ttFont):
    return build_filename(ttFont)


def inherit_vertical_metrics(ttFonts, family_name=None):
    """Inherit the vertical metrics from the same family which is
    hosted on Google Fonts.

    Args:
        ttFonts: a list of TTFont instances which belong to a family
        family_name: Optional string which allows users to specify a
            different family to inherit from e.g "Maven Pro".
    """
    family_name = font_familyname(ttFonts[0]) if not family_name else family_name

    gf_fonts = list(map(TTFont, download_family_from_Google_Fonts(family_name)))
    gf_fonts = {font_stylename(f): f for f in gf_fonts}
    # TODO (Marc F) use Regular font instead. If VF use font which has Regular
    # instance
    gf_fallback = list(gf_fonts.values())[0]

    fonts = {font_stylename(f): f for f in ttFonts}
    for style, font in fonts.items():
        if style in gf_fonts:
            src_font = gf_fonts[style]
        else:
            src_font = gf_fallback
        copy_vertical_metrics(src_font, font)

        if typo_metrics_enabled(src_font):
            font["OS/2"].fsSelection |= 1 << 7


def fix_vertical_metrics(ttFonts: Sequence[TTFont]):
    """Fix a family's vertical metrics based on:
    https://github.com/googlefonts/gf-docs/tree/main/VerticalMetrics

    Args:
        ttFonts: a list of TTFont instances which belong to a family
    """
    src_font = next((f for f in ttFonts if font_stylename(f) == "Regular"), ttFonts[0])

    # TODO (Marc F) CJK Fonts?

    # If OS/2.fsSelection bit 7 isn't enabled, enable it and set the typo metrics
    # to the previous win metrics.
    if not typo_metrics_enabled(src_font):
        src_font["OS/2"].fsSelection |= 1 << 7  # enable USE_TYPO_METRICS
        src_font["OS/2"].sTypoAscender = src_font["OS/2"].usWinAscent
        src_font["OS/2"].sTypoDescender = -src_font["OS/2"].usWinDescent
        src_font["OS/2"].sTypoLineGap = 0

    # Set the hhea metrics so they are the same as the typo
    src_font["hhea"].ascent = src_font["OS/2"].sTypoAscender
    src_font["hhea"].descent = src_font["OS/2"].sTypoDescender
    src_font["hhea"].lineGap = src_font["OS/2"].sTypoLineGap

    # Set the win Ascent and win Descent to match the family's bounding box
    win_desc, win_asc = family_bounding_box(ttFonts)
    src_font["OS/2"].usWinAscent = win_asc
    src_font["OS/2"].usWinDescent = abs(win_desc)

    # Set all fonts vertical metrics so they match the src_font
    for ttFont in ttFonts:
        ttFont["OS/2"].fsSelection |= 1 << 7
        copy_vertical_metrics(src_font, ttFont)


def copy_vertical_metrics(src_font, dst_font):
    for table, key in [
        ("OS/2", "usWinAscent"),
        ("OS/2", "usWinDescent"),
        ("OS/2", "sTypoAscender"),
        ("OS/2", "sTypoDescender"),
        ("OS/2", "sTypoLineGap"),
        ("hhea", "ascent"),
        ("hhea", "descent"),
        ("hhea", "lineGap"),
    ]:
        ratio = dst_font["head"].unitsPerEm / src_font["head"].unitsPerEm
        val = int(getattr(src_font[table], key) * ratio)
        setattr(dst_font[table], key, val)


def fix_italic_angle(ttFont) -> FixResult:
    style_name = font_stylename(ttFont)
    if "Italic" not in style_name:
        return _expect(ttFont, "post", "italicAngle", 0)
    if "Italic" in style_name and ttFont["hhea"].caretSlopeRun != 0:
        return _expect(
            ttFont,
            "post",
            "italicAngle",
            -math.degrees(
                math.atan(ttFont["hhea"].caretSlopeRun / ttFont["hhea"].caretSlopeRise)
            ),
        )
    return False, []


@fixes("com.google.fonts/check/caret_slope")
def fix_hhea_caret_slope_run(ttFont: TTFont) -> FixResult:
    if ttFont["post"].italicAngle == 0:
        return False, []

    expected_rise = ttFont["head"].unitsPerEm
    expected_run = round(
        math.tan(math.radians(-1 * ttFont["post"].italicAngle))
        * ttFont["head"].unitsPerEm
    )
    return _combine_results(
        _expect(ttFont, "hhea", "caretSlopeRise", expected_rise),
        _expect(ttFont, "hhea", "caretSlopeRun", expected_run),
    )


@fixes("com.google.fonts/check/name/unwanted_chars")
def fix_ascii_fontmetadata(font: TTFont) -> FixResult:
    """Fixes TTF 'name' table strings to be ascii only"""
    results = []
    for name in font["name"].names:
        old = name.string.decode(name.getEncoding())
        title = normalize_unicode_marks(old)
        if title != old:
            name.string = title.encode(name.getEncoding())
            results.append(f"Updated nameID={name.nameId} '{old}' to '{title}'")
    return font, results


def convert_cmap_subtables_to_v4(font):
    """Converts all cmap subtables to format 4.

    Returns a list of tuples (format, platformID, platEncID) of the tables
    which needed conversion."""
    cmap = font["cmap"]
    outtables = []
    converted = []
    for table in cmap.tables:
        if table.format != 4:
            converted.append((table.format, table.platformID, table.platEncID))
        newtable = CmapSubtable.newSubtable(4)
        newtable.platformID = table.platformID
        newtable.platEncID = table.platEncID
        newtable.language = table.language
        newtable.cmap = table.cmap
        outtables.append(newtable)
    font["cmap"].tables = outtables
    return converted


def drop_nonpid0_cmap(font, report=True):
    keep, drop = partition_cmap(font, lambda table: table.platformID == 0, report)
    return drop


def drop_mac_cmap(font, report=True):
    keep, drop = partition_cmap(
        font, lambda table: table.platformID != 1 or table.platEncID != 0, report
    )
    return drop


def fix_pua(font) -> FixResult:
    unencoded_glyphs = get_unencoded_glyphs(font)
    if not unencoded_glyphs:
        return False, []

    ucs2cmap = None
    cmap = font["cmap"]

    # Check if an UCS-2 cmap exists
    for ucs2cmapid in ((3, 1), (0, 3), (3, 0)):
        ucs2cmap = cmap.getcmap(ucs2cmapid[0], ucs2cmapid[1])
        if ucs2cmap:
            break
    # Create UCS-4 cmap and copy the contents of UCS-2 cmap
    # unless UCS 4 cmap already exists
    ucs4cmap = cmap.getcmap(3, 10)
    if not ucs4cmap:
        cmapModule = getTableModule("cmap")
        ucs4cmap = cmapModule.cmap_format_12(12)
        ucs4cmap.platformID = 3
        ucs4cmap.platEncID = 10
        ucs4cmap.language = 0
        if ucs2cmap:
            ucs4cmap.cmap = deepcopy(ucs2cmap.cmap)
        cmap.tables.append(ucs4cmap)
    # Map all glyphs to UCS-4 cmap Supplementary PUA-A codepoints
    # by 0xF0000 + glyphID
    ucs4cmap = cmap.getcmap(3, 10)
    for glyphID, glyph in enumerate(font.getGlyphOrder()):
        if glyph in unencoded_glyphs:
            ucs4cmap.cmap[0xF0000 + glyphID] = glyph
    font["cmap"] = cmap
    return True, ["Added UCS-4 cmap for PUA glyphs"]


@fixes("com.google.fonts/check/monospace")
def fix_isFixedPitch(ttfont) -> FixResult:
    same_width = set()
    glyph_metrics = ttfont["hmtx"].metrics
    messages = []
    for character in [chr(c) for c in range(65, 91)]:
        same_width.add(glyph_metrics[character][0])

    if len(same_width) == 1:
        _, messages = _expect(ttfont, "post", "isFixedPitch", 1)

        familyType = ttfont["OS/2"].panose.bFamilyType
        if familyType == 2:
            expected = 9
        elif familyType == 3 or familyType == 5:
            expected = 3
        elif familyType == 0:
            messages.append(
                "Font is monospace but panose fields seems to be not set."
                " Setting values to defaults (FamilyType = 2, Proportion = 9)."
            )
            ttfont["OS/2"].panose.bFamilyType = 2
            ttfont["OS/2"].panose.bProportion = 9
            expected = None
        else:
            expected = None

        if expected:
            if ttfont["OS/2"].panose.bProportion == expected:
                messages.append("Skipping OS/2.panose.bProportion is set correctly")
            else:
                messages.append(
                    (
                        "Font is monospace."
                        " Since OS/2.panose.bFamilyType is {}"
                        " we're updating OS/2.panose.bProportion"
                        " to {}"
                    ).format(familyType, expected)
                )
                ttfont["OS/2"].panose.bProportion = expected
                changed = True

        widths = [m[0] for m in ttfont["hmtx"].metrics.values() if m[0] > 0]
        width_max = max(widths)
        avg_width = otRound(sum(widths) / len(widths))
        changed, messages = _combine_results(
            (ttfont, messages),
            _expect(ttfont, "hhea", "advanceWidthMax", width_max),
            _expect(ttfont, "OS/2", "xAvgCharWidth", avg_width),
        )
    else:

        def set_panose(fnt, value):
            fnt["OS/2"].panose.bProportion = value

        changed, messages = _combine_results(
            _expect(ttfont, "post", "isFixedPitch", 0),
            _expect(
                ttfont,
                "OS/2",
                "panose.bProportion",
                0,
                getter=lambda fnt: fnt["OS/2"].panose.bProportion,
                setter=set_panose,
            ),
        )
    return changed, messages


def drop_superfluous_mac_names(ttfont) -> FixResult:
    """Drop superfluous Mac nameIDs.

    The following nameIDS are kept:
    1: Font Family name,
    2: Font Family Subfamily name,
    3: Unique font identifier,
    4: Full font name,
    5: Version string,
    6: Postscript name,
    16: Typographic family name,
    17: Typographic Subfamily name
    18: Compatible full (Macintosh only),
    20: PostScript CID,
    21: WWS Family Name,
    22: WWS Subfamily Name,
    25: Variations PostScript Name Prefix.

    We keep these IDs in order for certain application to still function
    such as Word 2011. IDs 1-6 are very common, > 16 are edge cases.

    https://www.microsoft.com/typography/otspec/name.htm"""
    drop_mac_names(ttfont, keep_ids=[1, 2, 3, 4, 5, 6, 16, 17, 18, 20, 21, 22, 25])


def drop_mac_names(ttfont, keep_ids=[]) -> FixResult:
    """Drop all mac names"""
    messages = []
    for namerecord in list(  # list() to avoid removing while iterating
        ttfont["name"].names
    ):
        if namerecord.nameID not in keep_ids:
            messages.append(f"Removed nameID {namerecord.nameID}: {namerecord.toStr()}")
            if (
                namerecord.platformID == 1
                and namerecord.platEncID == 0
                and namerecord.langID == 0
            ):
                ttfont["name"].names.remove(namerecord)
    return ttfont, messages


@fixes("com.google.fonts/check/empty_glyph_on_gid1_for_colrv0")
def fix_colr_v0_gid1(ttfont) -> FixResult:
    assert "COLR" in ttfont and ttfont["COLR"].version == 0
    if ttfont["maxp"].numGlyphs < 2:
        return ttfont, []
    glyph_names = ttfont.getGlyphOrder()
    glyf_table = ttfont["glyf"]
    second_glyph = glyph_names[1]
    if glyf_table[second_glyph].numberOfContours == 0:
        return ttfont, []
    has_empty_glyphs = any(glyf_table[g].numberOfContours == 0 for g in glyph_names)
    if has_empty_glyphs:
        return (
            _swap_empty_glyph_to_gid1(ttfont),
            ["Swapped empty glyph to GID 1 in COLR v0 font"],
        )
    else:
        return (
            _add_empty_glyph_to_gid1(ttfont),
            ["Added empty glyph to GID 1 in COLR v0 font"],
        )


def _swap_empty_glyph_to_gid1(ttfont):
    from nanoemoji.reorder_glyphs import reorder_glyphs

    glyf_table = ttfont["glyf"]
    ttfont.ensureDecompiled()
    glyph_order = ttfont.getGlyphOrder()
    empty_glyph = next(
        (g for g in glyph_order if glyf_table[g].numberOfContours == 0), None
    )
    if empty_glyph is None:
        raise ValueError(
            "Font contains no empty glyphs. Please include a space or .null glyph"
        )
    new_order = list(glyph_order)
    new_order.remove(empty_glyph)
    new_order.insert(1, empty_glyph)
    reorder_glyphs(ttfont, new_order)
    return True


def _add_empty_glyph_to_gid1(ttfont):
    from nanoemoji.util import load_fully
    from fontTools.ttLib.tables._g_l_y_f import Glyph
    from fontTools.ttLib.tables.otTables import NO_VARIATION_INDEX

    ttfont = load_fully(ttfont)
    glyph_order = ttfont.getGlyphOrder()
    glyf_table = ttfont["glyf"]
    hmtx = ttfont["hmtx"]

    empty_glyph = Glyph()
    empty_name = ".null" if ".null" not in glyph_order else "emptyglyph"
    assert empty_name not in glyph_order, f"{empty_name} already exists in font"
    glyf_table.glyphs[empty_name] = empty_glyph
    hmtx.metrics[empty_name] = (0, 0)
    if "HVAR" in ttfont:
        hvar = ttfont["HVAR"].table
        if hvar.AdvWidthMap:
            hvar.AdvWidthMap.mapping[empty_name] = NO_VARIATION_INDEX
        if hvar.LsbMap:
            hvar.LsbMap.mapping[empty_name] = NO_VARIATION_INDEX
        if hvar.RsbMap:
            hvar.RsbMap.mapping[empty_name] = NO_VARIATION_INDEX

    new_order = list(glyph_order)
    new_order.insert(1, empty_name)
    ttfont.setGlyphOrder(new_order)
    return True


@fixes("com.google.fonts/check/colorfont_tables")
def fix_colr_v1_add_svg(ttfont) -> FixResult:
    if "SVG " in ttfont:
        return ttfont, []
    font_filename = os.path.basename(ttfont.reader.file.name)
    with tempfile.TemporaryDirectory() as build_dir:
        subprocess.run(
            [
                "maximum_color",
                ttfont.reader.file.name,
                "--build_dir",
                build_dir,
                "--output_file",
                font_filename,
            ],
            check=True,
        )
        out_fp = os.path.join(build_dir, font_filename)
        fixed_ttfont = TTFont(out_fp)
        assert "SVG " in fixed_ttfont, "SVG table is missing"
        ttfont["SVG "] = fixed_ttfont["SVG "]
        return True, ["Added SVG table to COLR v1 font"]


@fixes(
    "com.google.fonts/check/colorfont_tables",
    "com.google.fonts/check/empty_glyph_on_gid1_for_colrv0",
)
def fix_colr_font(ttfont: TTFont) -> FixResult:
    """For COLR v0 fonts, we need to ensure that the 2nd glyph is whitespace glyph,
    https://github.com/googlefonts/gftools/issues/609. For COLR v1 fonts, we need
    to run Nanoemoji's maximum_color script in order to generate an SVG table for
    applications that don't support COLRv1 yet,
    https://github.com/googlefonts/fontbakery/issues/3888.
    """
    if "COLR" not in ttfont:
        return ttfont, []
    colr_version = ttfont["COLR"].version
    if colr_version == 0:
        return fix_colr_v0_gid1(ttfont)
    elif colr_version == 1:
        return fix_colr_v1_add_svg(ttfont)
    else:
        raise NotImplementedError(f"COLR version '{colr_version}' not supported.")


# def fix_nbspace_glyph(ttfont: TTFont):
#     cmap = ttfont.getBestCmap()
#     if 0x00A0 in cmap:
#         return


@fixes("com.google.fonts/check/name/license", "com.google.fonts/check/name/license_url")
def fix_license_strings(ttfont: TTFont) -> FixResult:
    """Update font's nametable license and license url strings"""
    from gftools.constants import OFL_LICENSE_URL, OFL_LICENSE_INFO

    name_table = ttfont["name"]
    for r in name_table.names:
        if r.nameID == 13:
            current_string = r.toUnicode()
            if "SIL Open Font License" in current_string:
                name_table.setName(
                    OFL_LICENSE_INFO, r.nameID, r.platformID, r.platEncID, r.langID
                )
                name_table.setName(
                    OFL_LICENSE_URL, 14, r.platformID, r.platEncID, r.langID
                )
                return True, ["Updated license strings"]
    return False, []


def fix_ofl_license(ttfont):
    """Generate the OFL.txt text from a font"""
    font_copyright = ttfont["name"].getName(0, 3, 1, 0x409)
    font_name = ttfont["name"].getName(1, 3, 1, 0x409)
    if not font_copyright or not font_name:
        raise ValueError("Font doesn't contain copyright or name strings")
    font_copyright = font_copyright.toUnicode()
    if "reserved font name" in font_copyright.lower():
        raise ValueError(
            "Font copyright has Reserved Font Name. Please ask if it can be removed."
        )
    font_name = font_name.toUnicode()
    font_year = re.search(r"[0-9]{4}", font_copyright)
    font_year = datetime.now().year if not font_year else font_year.group(0)
    git_url = re.search(r"(\()(.*)(\))", font_copyright)
    if not git_url:
        raise ValueError("Font copyright doesn't contain a git url")
    first_line = (
        f"Copyright {font_year} The {font_name} Project Authors ({git_url.group(2)})"
    )
    return f"{first_line}\n{OFL_BODY_TEXT}"


@fixes("com.google.fonts/check/metadata/valid_nameid25")
def fix_no_varpsname(ttFont: TTFont) -> FixResult:
    if "fvar" not in ttFont:
        return ttFont, []
    name_table = ttFont["name"]
    variation_ps_name = name_table.getName(25, 3, 1, 0x409)
    if not variation_ps_name:
        build_variations_ps_name(ttFont)
        var_ps_name = ttFont["name"].getName(25, 3, 1, 0x409).toUnicode()
        return ttFont, [
            f"Added a Variations PostScript Name Prefix (NameID 25) '{var_ps_name}'"
        ]
    return ttFont, []


def fix_font(
    font: TTFont,
    include_source_fixes: bool = False,
    new_family_name: str | None = None,
    fvar_instance_axis_dflts: dict[str, float] | None = None,
    overwrite_fvar_instances: bool = True,
) -> TTFont:
    fixed_font = deepcopy(font)
    if new_family_name:
        rename_font(fixed_font, new_family_name)
    if fixed_font["OS/2"].version > 1:
        fixed_font["OS/2"].version = 4

    fixes = [
        fix_license_strings,
        fix_hinted_font,
        fix_unhinted_font,
        fix_no_varpsname,
        fix_colr_font,
        fix_hhea_caret_slope_run,
    ]

    if include_source_fixes:
        fixes.extend(
            [
                remove_tables,
                fix_nametable,
                fix_fs_type,
                fix_fs_selection,
                fix_mac_style,
                fix_weight_class,
                fix_italic_angle,
            ]
        )

    for fixer in fixes:
        result = fixer(fixed_font)
        if len(result) != 2:
            log.error(f"{fixer} returned an unexpected result: {result}")
        _, messages = result
        if messages:
            log.info("\n".join(messages))

    if overwrite_fvar_instances:
        fix_fvar_instances(fixed_font, fvar_instance_axis_dflts)
    return fixed_font


def fix_family(
    fonts,
    include_source_fixes=False,
    new_family_name=None,
    fvar_instance_axis_dflts=None,
):
    """Fix all fonts in a family"""
    validate_family(fonts)

    fixed_fonts = []
    for font in fonts:
        fixed_font = fix_font(
            font,
            include_source_fixes=include_source_fixes,
            new_family_name=new_family_name,
            fvar_instance_axis_dflts=fvar_instance_axis_dflts,
        )
        fixed_fonts.append(fixed_font)
    family_name = font_familyname(fixed_fonts[0])
    if include_source_fixes:
        try:
            if Google_Fonts_has_family(family_name):
                inherit_vertical_metrics(fixed_fonts)
            else:
                log.warning(
                    f"{family_name} is not on Google Fonts. Skipping "
                    "regression fixes"
                )
        except FileNotFoundError:
            log.warning(
                f"Google Fonts api key not found so we can't regression "
                "fix fonts. See Repo readme to add keys."
            )
        fix_vertical_metrics(fixed_fonts)
        if all(["fvar" in f for f in fonts]):
            gen_stat_tables(fixed_fonts)
    return fixed_fonts


class FontFixer:
    def __init__(self, path, report=True, verbose=False, **kwargs):
        self.font = TTFont(path)
        self.path = path
        self.font_filename = basename(path)
        self.saveit = False
        self.report = report
        self.verbose = verbose
        self.messages = []
        self.args = kwargs
        self.fixes = []
        if "fixes" in kwargs:
            self.fixes = kwargs["fixes"]

    def __del__(self):
        if self.report:
            print("\n".join(self.messages))
        if self.saveit:
            if self.verbose:
                print("Saving %s to %s.fix" % (self.font_filename, self.path))
            self.font.save(self.path + ".fix")
        elif self.verbose:
            print("There were no changes needed on %s!" % self.font_filename)

    def show(self):
        pass

    def fix(self):
        for f in self.fixes:
            rv = f(self.font)
            if isinstance(rv, tuple) and len(rv) == 2:
                changed, messages = rv
                self.messages.extend(messages)
            else:
                changed = rv
            if changed:
                self.saveit = True


class GaspFixer(FontFixer):
    def fix(self, value=15):
        try:
            table = self.font.get("gasp")
            table.gaspRange[65535] = value
            self.saveit = True
        except:
            print(
                ("ER: {}: no table gasp... " "Creating new table. ").format(self.path)
            )
            table = ttLib.newTable("gasp")
            table.gaspRange = {65535: value}
            self.font["gasp"] = table
            self.saveit = True

    def show(self):
        try:
            self.font.get("gasp")
        except:
            print("ER: {}: no table gasp".format(self.path))
            return

        try:
            print(self.font.get("gasp").gaspRange[65535])
        except IndexError:
            print("ER: {}: no index 65535".format(self.path))



#!/usr/bin/env python3
#
# Copyright 2017 The Google Font Tools Authors
# Copyright 2016 The Fontbakery Authors
#
# Copyright 2010, Google Inc.
# Author: Dave Crossland (dave@understandinglimited.com)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
A Python script for printing bounding boxes to stdout.

Users can either check a collection of fonts bounding boxes (--family) or
the bounding box for each glyph in the collection of fonts (--glyphs).
"""
from argparse import ArgumentParser, RawTextHelpFormatter
import csv
import sys
from fontTools.ttLib import TTFont
import tabulate

parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
parser.add_argument("fonts", nargs="+", help="Fonts in OpenType (TTF/OTF) format")
parser.add_argument(
    "--csv",
    default=False,
    action="store_true",
    help="Output data in comma-separated-values format",
)
parser.add_argument(
    "--extremes",
    default=False,
    action="store_true",
    help="Print extremes coordinates for each category",
)
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument(
    "--glyphs",
    default=False,
    action="store_true",
    help=("Return the bounds for glyphs" " in a collection of fonts"),
)
group.add_argument(
    "--family",
    default=False,
    action="store_true",
    help="Return the bounds for a family of fonts",
)


def printInfo(rows, save=False):
    header = [r[0] for r in rows[0]]
    t = []
    for row in rows:
        t.append([r[1] for r in row])

    if save:
        writer = csv.writer(sys.stdout)
        writer.writerows([header])
        writer.writerows(t)
        sys.exit(0)
    else:
        print(tabulate.tabulate(t, header, tablefmt="pipe"))


def find_extremes(rows):
    extremes = {}

    for row in rows:
        for k, v in row:
            if type(v) == str:
                continue
            if k not in extremes:
                extremes[k] = int(v)
            else:
                if abs(int(v)) > abs(extremes[k]):
                    extremes[k] = v
    return [extremes.items()]


def main(args=None):
    args = parser.parse_args(args)

    rows = []
    for font in args.fonts:
        font_path = font
        font = TTFont(font_path)
        if args.glyphs:
            for g_name in font["glyf"].glyphs:
                glyph = font["glyf"][g_name]
                try:
                    rows.append(
                        [
                            ("Font", font_path),
                            ("Glyph", g_name),
                            ("xMin", glyph.xMin),
                            ("yMin", glyph.yMin),
                            ("xMax", glyph.xMax),
                            ("yMax", glyph.yMax),
                        ]
                    )
                except AttributeError:
                    # glyphs without paths or components don't have
                    # yMin, yMax etc
                    rows.append(
                        [
                            ("Font", font_path),
                            ("Glyph", g_name),
                            ("xMin", 0),
                            ("yMin", 0),
                            ("xMax", 0),
                            ("yMax", 0),
                        ]
                    )
                    pass

        elif args.family:
            rows.append(
                [
                    ("Font", font_path),
                    ("xMin", font["head"].xMin),
                    ("yMin", font["head"].yMin),
                    ("xMax", font["head"].xMax),
                    ("yMax", font["head"].yMax),
                ]
            )

    if args.extremes:
        rows = find_extremes(rows)

    if args.csv:
        printInfo(rows, save=True)
    else:
        printInfo(rows)


if __name__ == "__main__":
    main()



"""Render strings from a font as a waterfall PNG.

The rendering backend is platform-native by default:
    macOS   -> CoreText
    Windows -> DirectWrite (implemented via Skia-Python; see
               directwrite_backend.py for rationale)
    Linux   -> FreeType (with HarfBuzz for shaping)

See ``docs/gftools-render-text/spec.md`` for the design spec.
"""

from __future__ import annotations

import platform
import re
import sys
from pathlib import Path
from typing import Iterable, Iterator

from fontTools.ttLib import TTFont
from PIL import Image, ImageChops, ImageDraw, ImageFont


DEFAULT_PPEMS: tuple[int, ...] = (8, 10, 12, 14, 16, 20, 24, 36)
ROW_PADDING = 6
CANVAS_PADDING = 12
# Per-row vertical padding above ascender and below descender. Shared by all
# backends so row heights match exactly across platforms.
ROW_VPAD = 2


def default_backend() -> str:
    if sys.platform == "darwin":
        return "coretext"
    if sys.platform == "win32":
        return "directwrite"
    return "freetype"


def parse_variations(spec: str) -> dict[str, float]:
    """Parse ``wght=400,wdth=75`` into ``{"wght": 400.0, "wdth": 75.0}``."""
    result: dict[str, float] = {}
    for token in spec.split(","):
        token = token.strip()
        if not token:
            continue
        if "=" not in token:
            raise ValueError(f"variation token {token!r} missing '='")
        axis, value = token.split("=", 1)
        axis = axis.strip()
        if len(axis) != 4:
            raise ValueError(f"axis tag must be 4 chars, got {axis!r}")
        result[axis] = float(value.strip())
    return result


def is_variable(font_path: Path) -> bool:
    with TTFont(font_path) as font:
        return "fvar" in font


def iter_fvar_instances(
    font_path: Path,
) -> Iterator[tuple[str, dict[str, float]]]:
    """Yield ``(subfamily_name, {axis_tag: value})`` for each fvar instance."""
    with TTFont(font_path) as font:
        if "fvar" not in font:
            return
        fvar = font["fvar"]
        name = font["name"]
        for inst in fvar.instances:
            label = name.getDebugName(inst.subfamilyNameID) or "Instance"
            yield label, dict(inst.coordinates)


_FILENAME_SAFE_RE = re.compile(r"[^A-Za-z0-9._-]+")


def _filename_safe(s: str) -> str:
    return _FILENAME_SAFE_RE.sub("", s)


def _format_value(value: float) -> str:
    return str(int(value)) if float(value).is_integer() else f"{value:g}"


def _format_variations_suffix(variations: dict[str, float]) -> str:
    return "-".join(f"{tag}{_format_value(val)}" for tag, val in variations.items())


def output_path_for(
    font_path: Path,
    *,
    variations: dict[str, float] | None = None,
    output: str | Path | None = None,
) -> Path:
    if output is not None:
        return Path(output)
    stem = font_path.stem
    if variations:
        stem = f"{stem}-{_format_variations_suffix(variations)}"
    return font_path.parent / f"{stem}.png"


def output_dir_for_all(
    font_path: Path, *, output_dir: str | Path | None = None
) -> Path:
    if output_dir is not None:
        return Path(output_dir)
    return font_path.parent / f"{font_path.stem}_imgs"


def output_dir_for_diff(
    after_font_path: Path, *, output_dir: str | Path | None = None
) -> Path:
    if output_dir is not None:
        return Path(output_dir)
    return after_font_path.parent / f"{after_font_path.stem}_diff"


def output_path_for_instance(
    font_path: Path, instance_name: str, output_dir: Path
) -> Path:
    return output_dir / f"{font_path.stem}-{_filename_safe(instance_name)}.png"


def output_subdir_for_instance(out_dir: Path, instance_name: str) -> Path:
    """Return a per-instance subdirectory inside ``out_dir`` (used by ``diff --all``)."""
    return out_dir / _filename_safe(instance_name)


def render_waterfall(
    font_path: Path,
    text: str,
    *,
    ppems: Iterable[int] = DEFAULT_PPEMS,
    variations: dict[str, float] | None = None,
    backend: str | None = None,
) -> Image.Image:
    """Render ``text`` from ``font_path`` at each ppem and stack vertically."""
    backend = backend or default_backend()
    render_row = _load_backend(backend)
    ascent_du, descent_du, upem = _font_line_metrics(font_path)
    rows = []
    for ppem in ppems:
        ascent_px = int(round(ascent_du * ppem / upem))
        descent_px = int(round(descent_du * ppem / upem))
        target_h = ascent_px + descent_px + ROW_VPAD * 2
        baseline_y = ascent_px + ROW_VPAD
        rows.append(
            render_row(
                font_path,
                text,
                ppem,
                variations,
                target_height=target_h,
                baseline_y=baseline_y,
            )
        )
    canvas = _compose_waterfall(rows)
    _annotate_platform(canvas, backend)
    return canvas


def _font_line_metrics(font_path: Path) -> tuple[int, int, int]:
    """Return ``(ascent_du, descent_du, upem)`` from OS/2 typo metrics.

    All backends use this single source so row heights match across platforms.
    Falls back to ``hhea`` for fonts without an OS/2 table.
    """
    with TTFont(font_path) as font:
        upem = font["head"].unitsPerEm
        os2 = font.get("OS/2")
        if os2 is not None:
            return os2.sTypoAscender, -os2.sTypoDescender, upem
        hhea = font["hhea"]
        return hhea.ascender, -hhea.descender, upem


_BACKEND_DISPLAY = {
    "coretext": "CoreText",
    "directwrite": "DirectWrite",
    "freetype": "FreeType",
}


def _annotate_platform(canvas: Image.Image, backend: str) -> None:
    label = f"{platform.system()} / {_BACKEND_DISPLAY.get(backend, backend)}"
    font = ImageFont.load_default(size=12)
    draw = ImageDraw.Draw(canvas)
    bbox = draw.textbbox((0, 0), label, font=font)
    text_height = bbox[3] - bbox[1]
    x = CANVAS_PADDING // 2
    y = canvas.height - text_height - CANVAS_PADDING // 2
    draw.text((x, y), label, fill=(128, 128, 128), font=font)


def _load_backend(name: str):
    if name == "freetype":
        from . import freetype_backend

        return freetype_backend.render_row
    if name == "coretext":
        from . import coretext_backend

        return coretext_backend.render_row
    if name == "directwrite":
        from . import directwrite_backend

        return directwrite_backend.render_row
    raise ValueError(f"unknown backend {name!r}")


def _compose_waterfall(rows: list[Image.Image]) -> Image.Image:
    width = max((row.width for row in rows), default=0) + CANVAS_PADDING * 2
    height = (
        sum(row.height for row in rows)
        + ROW_PADDING * (len(rows) - 1 if rows else 0)
        + CANVAS_PADDING * 2
    )
    canvas = Image.new("RGB", (width, height), "white")
    y = CANVAS_PADDING
    for row in rows:
        canvas.paste(row, (CANVAS_PADDING, y))
        y += row.height + ROW_PADDING
    return canvas


def pad_to_match(images: list[Image.Image]) -> list[Image.Image]:
    """Pad each image with white to the max (width, height) of the set."""
    if not images:
        return []
    w = max(im.width for im in images)
    h = max(im.height for im in images)
    return [_pad_white(im, w, h) for im in images]


def _pad_white(im: Image.Image, w: int, h: int) -> Image.Image:
    if im.size == (w, h):
        return im
    canvas = Image.new(im.mode, (w, h), "white")
    canvas.paste(im, (0, 0))
    return canvas


def diff_image(before: Image.Image, after: Image.Image) -> Image.Image:
    """Return the absolute per-pixel difference (PIL's 'difference' blend mode).

    The two inputs must be the same size; use :func:`pad_to_match` first.
    Identical pixels become black; differing pixels become brighter.
    """
    if before.size != after.size:
        raise ValueError(
            f"diff inputs must match in size; got {before.size} vs {after.size}"
        )
    return ImageChops.difference(after, before)


def save_animation(
    frames: list[Image.Image],
    path: Path,
    duration_ms: int = 500,
    labels: list[str] | None = None,
) -> None:
    """Save an animated GIF cycling through ``frames`` (infinite loop).

    If ``labels`` is given (one per frame), each frame is stamped with its
    label in the top-right corner so the active frame is identifiable
    while the GIF plays.
    """
    if not frames:
        raise ValueError("no frames to animate")
    if labels is not None:
        if len(labels) != len(frames):
            raise ValueError("labels must be the same length as frames")
        frames = [_stamp_top_right(f, label) for f, label in zip(frames, labels)]
    frames[0].save(
        path,
        save_all=True,
        append_images=frames[1:],
        duration=duration_ms,
        loop=0,
    )


def _stamp_top_right(im: Image.Image, label: str) -> Image.Image:
    out = im.copy()
    font = ImageFont.load_default(size=14)
    draw = ImageDraw.Draw(out)
    bbox = draw.textbbox((0, 0), label, font=font)
    text_w = bbox[2] - bbox[0]
    pad = 8
    x = out.width - text_w - pad
    y = pad
    draw.text((x, y), label, fill=(180, 30, 30), font=font)
    return out



from gftools.builder.operations import OperationBase
import glyphsLib
import gftools.builder


class PaintCompiler(OperationBase):
    description = "Run paintcompiler on a variable font"
    rule = "paintcompiler $args -o $out $in"

    @property
    def variables(self):
        vars = super().variables
        vars["args"] = self.original.get("args", "")
        return vars



#!/usr/bin/env python3
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
r"""Tool to identify problems with fonts.

"""
from __future__ import print_function
import collections
import contextlib
import os
import re
import sys

from fontTools import ttLib
from absl import flags, app
from gftools.util import google_fonts as fonts

FLAGS = flags.FLAGS

flags.DEFINE_boolean("suppress_pass", True, "Whether to print pass: results")
flags.DEFINE_boolean("check_metadata", True, "Whether to check METADATA values")
flags.DEFINE_boolean("check_font", True, "Whether to check font values")
flags.DEFINE_string("repair_script", None, "Where to write a repair script")
_FIX_TYPE_OPTS = [
    "all",
    "name",
    "filename",
    "postScriptName",
    "fullName",
    "fsSelection",
    "fsType",
    "usWeightClass",
    "emptyGlyphLSB",
]
flags.DEFINE_multi_string(
    "fix_type",
    "all",
    "What types of problems should be fixed by "
    "repair_script. "
    "Choices: " + ", ".join(_FIX_TYPE_OPTS),
)

ResultMessageTuple = collections.namedtuple(
    "ResultMessageTuple", ["happy", "message", "path", "repair_script"]
)


def _HappyResult(message, path):
    return ResultMessageTuple(True, message, path, None)


def _SadResult(message, path, repair_script=None):
    return ResultMessageTuple(False, message, path, repair_script)


def _DropEmptyPathSegments(path):
    """Removes empty segments from the end of path.

    Args:
      path: A filesystem path.
    Returns:
      path with trailing empty segments removed. Eg /duck/// => /duck.
    """
    while True:
        (head, tail) = os.path.split(path)
        if tail:
            break
        path = head
    return path


def _SanityCheck(path):
    """Runs various sanity checks on the font family under path.

    Args:
      path: A directory containing a METADATA.pb file.
    Returns:
      A list of ResultMessageTuple's.
    """
    try:
        fonts.Metadata(path)
    except ValueError as e:
        return [_SadResult("Bad METADATA.pb: " + e.message, path)]

    results = []
    if FLAGS.check_metadata:
        results.extend(_CheckLicense(path))
        results.extend(_CheckNameMatching(path))

    if FLAGS.check_font:
        results.extend(_CheckFontInternalValues(path))

    return results


def _CheckLicense(path):
    """Verifies that METADATA.pb license is correct under path.

    Assumes path is of the form .../<license>/<whatever>/METADATA.pb.

    Args:
      path: A directory containing a METADATA.pb file.
    Returns:
      A list with one ResultMessageTuple. If happy, license is good.
    """
    metadata = fonts.Metadata(path)
    lic = metadata.license
    lic_dir = os.path.basename(os.path.dirname(path))

    # We use /apache for the license Apache2
    if lic_dir == "apache":
        lic_dir += "2"

    result = _HappyResult("License consistantly %s" % lic, path)
    # if we were Python 3 we'd use casefold(); this will suffice
    if lic_dir.lower() != lic.lower():
        result = _SadResult(
            "Dir license != METADATA license: %s != %s" % (lic_dir, lic), path
        )

    return [result]


def _CheckNameMatching(path):
    """Verifies the various name fields in the METADATA.pb file are sane.

    Args:
      path: A directory containing a METADATA.pb file.
    Returns:
      A list of ResultMessageTuple, one per validation performed.
    """
    results = []
    metadata = fonts.Metadata(path)
    name = metadata.name

    for font in metadata.fonts:
        # We assume style/weight is correct in METADATA
        style = font.style
        weight = font.weight
        values = [
            ("name", name, font.name),
            ("filename", fonts.FilenameFor(name, style, weight, ".ttf"), font.filename),
            (
                "postScriptName",
                fonts.FilenameFor(name, style, weight),
                font.post_script_name,
            ),
            ("fullName", fonts.FullnameFor(name, style, weight), font.full_name),
        ]

        for key, expected, actual in values:
            if expected != actual:
                results.append(
                    _SadResult(
                        "%s METADATA %s/%d %s expected %s, got %s"
                        % (name, style, weight, key, expected, actual),
                        path,
                        _FixMetadata(style, weight, key, expected),
                    )
                )

    if not results:
        results.append(
            _HappyResult('METADATA name consistently derived from "%s"' % name, path)
        )

    return results


def _IsItalic(style):
    return style.lower() == "italic"


def _IsBold(weight):
    """Is this weight considered bold?

    Per Dave C, only 700 will be considered bold.

    Args:
      weight: Font weight.
    Returns:
      True if weight is considered bold, otherwise False.
    """
    return weight == 700


def _ShouldFix(key):
    return FLAGS.fix_type and (key in FLAGS.fix_type or "all" in FLAGS.fix_type)


def _FixMetadata(style, weight, key, expected):
    if not _ShouldFix(key):
        return None

    if not isinstance(expected, int):
        expected = "'%s'" % expected

    return (
        "[f for f in metadata.fonts if f.style == '%s' "
        "and f.weight == %d][0].%s = %s"
    ) % (style, weight, re.sub("([a-z])([A-Z])", r"\1_\2", key).lower(), expected)


def _FixFsSelectionBit(key, expected):
    """Write a repair script to fix a bad fsSelection bit.

    Args:
      key: The name of an fsSelection flag, eg 'ITALIC' or 'BOLD'.
      expected: Expected value, true/false, of the flag.
    Returns:
      A python script to fix the problem.
    """
    if not _ShouldFix("fsSelection"):
        return None

    op = "|="
    verb = "set"
    mask = bin(fonts.FsSelectionMask(key))
    if not expected:
        op = "&="
        verb = "unset"
        mask = "~" + mask

    return "ttf['OS/2'].fsSelection %s %s  # %s %s" % (op, mask, verb, key)


def _FixFsType(expected):
    if not _ShouldFix("fsType"):
        return None
    return "ttf['OS/2'].fsType = %d" % expected


def _FixWeightClass(expected):
    if not _ShouldFix("usWeightClass"):
        return None
    return "ttf['OS/2'].usWeightClass = %d" % expected


def _FixBadNameRecord(friendly_name, name_id, expected):
    if not _ShouldFix(friendly_name):
        return None

    return (
        "for nr in [n for n in ttf['name'].names if n.nameID == %d]:\n"
        "  nr.string = '%s'.encode(nr.getEncoding())  # Fix %s"
        % (name_id, expected, friendly_name)
    )


def _FixMissingNameRecord(friendly_name, name_id, expected):
    if not _ShouldFix(friendly_name):
        return None

    return (
        "nr = ttLib.tables._n_a_m_e.NameRecord()\n"
        "nr.nameID = %d  # %s"
        "nr.langID = 0x409\n"
        "nr.platEncID = 1\n"
        "nr.platformID = 3\n"
        "nr.string = '%s'.encode(nr.getEncoding())\n"
        "ttf['name'].names.append(nr)\n" % (name_id, friendly_name, expected)
    )


def _FixEmptyGlyphLsb(glyph_name):
    if not _ShouldFix("emptyGlyphLSB"):
        return None

    return "ttf['hmtx']['%s'] = [ttf['hmtx']['%s'][0], 0]\n" % (glyph_name, glyph_name)


def _CheckFontOS2Values(path, font, ttf):
    """Check sanity of values hidden in the 'OS/2' table.

    Notably usWeightClass, fsType, fsSelection.

    Args:
      path: Path to directory containing font.
      font: A font record from a METADATA.pb.
      ttf: A fontTools.ttLib.TTFont for the font.
    Returns:
      A list of ResultMessageTuple for tests performed.
    """
    results = []

    font_file = font.filename
    full_font_file = os.path.join(path, font_file)
    expected_style = font.style
    expected_weight = font.weight

    os2 = ttf["OS/2"]
    fs_selection_flags = fonts.FsSelectionFlags(os2.fsSelection)
    actual_weight = os2.usWeightClass
    fs_type = os2.fsType

    marked_oblique = "OBLIQUE" in fs_selection_flags
    marked_italic = "ITALIC" in fs_selection_flags
    marked_bold = "BOLD" in fs_selection_flags

    expect_italic = _IsItalic(expected_style)
    expect_bold = _IsBold(expected_weight)
    # Per Dave C, we should NEVER set oblique, use 0 for italic
    expect_oblique = False

    results.append(
        ResultMessageTuple(
            marked_italic == expect_italic,
            "%s %s/%d fsSelection marked_italic %d"
            % (font_file, expected_style, expected_weight, marked_italic),
            full_font_file,
            _FixFsSelectionBit("ITALIC", expect_italic),
        )
    )
    results.append(
        ResultMessageTuple(
            marked_bold == expect_bold,
            "%s %s/%d fsSelection marked_bold %d"
            % (font_file, expected_style, expected_weight, marked_bold),
            full_font_file,
            _FixFsSelectionBit("BOLD", expect_bold),
        )
    )

    results.append(
        ResultMessageTuple(
            marked_oblique == expect_oblique,
            "%s %s/%d fsSelection marked_oblique %d"
            % (font_file, expected_style, expected_weight, marked_oblique),
            full_font_file,
            _FixFsSelectionBit("OBLIQUE", expect_oblique),
        )
    )

    # For weight < 300, just confirm weight [250, 300)
    # TODO(user): we should also verify ordering is correct
    weight_ok = expected_weight == actual_weight
    weight_msg = str(expected_weight)
    if expected_weight < 300:
        weight_ok = actual_weight >= 250 and actual_weight < 300
        weight_msg = "[250, 300)"

    results.append(
        ResultMessageTuple(
            weight_ok,
            "%s %s/%d weight expected: %s usWeightClass: %d"
            % (font_file, expected_style, expected_weight, weight_msg, actual_weight),
            full_font_file,
            _FixWeightClass(expected_weight),
        )
    )

    expected_fs_type = 0
    results.append(
        ResultMessageTuple(
            expected_fs_type == fs_type,
            "%s %s/%d fsType expected: %d fsType: %d"
            % (font_file, expected_style, expected_weight, expected_fs_type, fs_type),
            full_font_file,
            _FixFsType(expected_fs_type),
        )
    )

    return results


def _CheckFontNameValues(path, name, font, ttf):
    """Check sanity of values in the 'name' table.

    Specifically the fullname and postScriptName.

    Args:
      path: Path to directory containing font.
      name: The name of the family.
      font: A font record from a METADATA.pb.
      ttf: A fontTools.ttLib.TTFont for the font.
    Returns:
      A list of ResultMessageTuple for tests performed.
    """
    results = []

    style = font.style
    weight = font.weight
    full_font_file = os.path.join(path, font.filename)

    expectations = [
        ("family", fonts.NAME_FAMILY, name),
        ("postScriptName", fonts.NAME_PSNAME, fonts.FilenameFor(name, style, weight)),
        ("fullName", fonts.NAME_FULLNAME, fonts.FullnameFor(name, style, weight)),
    ]

    for friendly_name, name_id, expected in expectations:
        # If you have lots of name records they should ALL have the right value
        actuals = fonts.ExtractNames(ttf, name_id)
        for idx, actual in enumerate(actuals):
            results.append(
                ResultMessageTuple(
                    expected == actual,
                    "%s %s/%d 'name' %s[%d] expected %s, got %s"
                    % (name, style, weight, friendly_name, idx, expected, actual),
                    full_font_file,
                    _FixBadNameRecord(friendly_name, name_id, expected),
                )
            )

        # should have at least one actual
        if not actuals:
            results.append(
                _SadResult(
                    "%s %s/%d 'name' %s has NO values"
                    % (name, style, weight, friendly_name),
                    full_font_file,
                    _FixMissingNameRecord(friendly_name, name_id, expected),
                )
            )

    return results


def _CheckLSB0ForEmptyGlyphs(path, font, ttf):
    """Checks if font has empty (loca[n] == loca[n+1]) glyphs that have non-0 lsb.

    There is no reason to set such lsb's.

    Args:
      path: Path to directory containing font.
      font: A font record from a METADATA.pb.
      ttf: A fontTools.ttLib.TTFont for the font.
    Returns:
      A list of ResultMessageTuple for tests performed.
    """
    results = []
    if "loca" not in ttf:
        return results
    for glyph_index, glyph_name in enumerate(ttf.getGlyphOrder()):
        is_empty = ttf["loca"][glyph_index] == ttf["loca"][glyph_index + 1]
        lsb = ttf["hmtx"][glyph_name][1]
        if is_empty and lsb != 0:
            results.append(
                _SadResult(
                    "%s %s/%d ['hmtx']['%s'][1] (lsb) should be 0 but is %d"
                    % (font.name, font.style, font.weight, glyph_name, lsb),
                    os.path.join(path, font.filename),
                    _FixEmptyGlyphLsb(glyph_name),
                )
            )
    return results


def _CheckFontInternalValues(path):
    """Validates fonts internal metadata matches METADATA.pb values.

    In particular, checks 'OS/2' {usWeightClass, fsSelection, fsType} and 'name'
    {fullName, postScriptName} values.

    Args:
      path: A directory containing a METADATA.pb file.
    Returns:
      A list of ResultMessageTuple, one per validation performed.
    """
    results = []
    metadata = fonts.Metadata(path)
    name = metadata.name

    for font in metadata.fonts:
        font_file = font.filename
        with contextlib.closing(ttLib.TTFont(os.path.join(path, font_file))) as ttf:
            results.extend(_CheckFontOS2Values(path, font, ttf))
            results.extend(_CheckFontNameValues(path, name, font, ttf))
            results.extend(_CheckLSB0ForEmptyGlyphs(path, font, ttf))

    return results


def _WriteRepairScript(dest_file, results):
    with open(dest_file, "w") as out:
        out.write("import collections\n")
        out.write("import contextlib\n")
        out.write("from fontTools import ttLib\n")
        out.write("from google.protobuf.text_format " "import text_format\n")
        out.write("from gftools.fonts_public_pb2 import fonts_pb2\n")
        out.write("from gftools.fonts_public_pb2 " "import fonts_metadata_pb2\n")
        out.write("\n")

        # group by path
        by_path = collections.defaultdict(list)
        for result in results:
            if result.happy or not result.repair_script:
                continue
            if result.repair_script not in by_path[result.path]:
                by_path[result.path].append(result.repair_script)

        for path in sorted(by_path.keys()):
            out.write("# repair %s\n" % os.path.basename(path))
            _, ext = os.path.splitext(path)

            prefix = ""
            if ext == ".ttf":
                prefix = "  "
                out.write(
                    "with contextlib.closing(ttLib.TTFont('%s')) as ttf:\n" % path
                )
            elif os.path.isdir(path):
                metadata_file = os.path.join(path, "METADATA.pb")
                out.write("metadata = fonts_pb2.FamilyProto()\n")
                out.write("with open('%s', 'r') as f:\n" % metadata_file)
                out.write("  text_format.Merge(f.read(), metadata)\n")
            else:
                raise ValueError("Not sure how to script %s" % path)

            for repair in by_path[path]:
                out.write(prefix)
                out.write(re.sub("\n", "\n" + prefix, repair))
                out.write("\n")

            if ext == ".ttf":
                out.write("  ttf.save('%s')\n" % path)

            if os.path.isdir(path):
                out.write("with open('%s', 'w') as f:\n" % metadata_file)
                out.write("  f.write(text_format.MessageToString(metadata))\n")

            out.write("\n")


def main(argv):
    result_code = 0
    all_results = []
    paths = [_DropEmptyPathSegments(os.path.expanduser(p)) for p in argv[1:]]
    for path in paths:
        if not os.path.isdir(path):
            raise ValueError("Not a directory: %s" % path)

    for path in paths:
        for font_dir in fonts.FontDirs(path):
            results = _SanityCheck(font_dir)
            all_results.extend(results)
            for result in results:
                result_msg = "pass"
                if not result.happy:
                    result_code = 1
                    result_msg = "FAIL"
                if not result.happy or not FLAGS.suppress_pass:
                    print("%s: %s (%s)" % (result_msg, result.message, font_dir))

    if FLAGS.repair_script:
        _WriteRepairScript(FLAGS.repair_script, all_results)

    sys.exit(result_code)


if __name__ == "__main__":
    app.run(main)



"""
gftools gen-avar2

Generate an avar2 table for each font in a variable font family
from a yaml file.
```
MavenPro[wght].ttf:
- in:
    wght: 400.0
  out:
    wght: 400.0
- in:
    wght: 900.0
  out:
    wght: 900.0
```
"""

import argparse
import os
from collections import OrderedDict
from fontTools.ttLib import TTFont
from fontTools.varLib import _add_avar
from fontTools.designspaceLib import AxisMappingDescriptor, AxisDescriptor
import yaml
from fontTools.varLib.avar.unbuild import _denormalize
from fontTools.ttLib.tables._f_v_a_r import Axis
from fontTools.misc.textTools import Tag
from gftools.fix import fix_fvar_instances
from fontTools.ttLib.tables import otTables


def gen_fvar_axes(font, mapping):
    """A mapping may declare axes such as wght, wdth etc that may not exist
    in the font as a set of masters."""
    # TODO (M Foley) improve and move this logic to fonttools.
    axes_to_make = {}
    nameTable = font["name"]
    axis_count = len(font["fvar"].axes)
    for m in mapping:
        for axis in m["in"]:
            if axis not in axes_to_make:
                axis_count += 1
                axes_to_make[axis] = {"min": m["in"][axis], "max": m["in"][axis]}
            axes_to_make[axis]["min"] = min(axes_to_make[axis]["min"], m["in"][axis])
            axes_to_make[axis]["max"] = max(axes_to_make[axis]["max"], m["in"][axis])

    # Update the gvar axis count before adding new axes in order to clear an assertion
    # in fonttools
    font["gvar"].axisCount = axis_count
    for axis_name, axis_range in axes_to_make.items():
        if axis_name not in font["fvar"].axes:
            axis = Axis()
            axis.axisTag = Tag(axis_name)
            axis.minValue, axis.defaultValue, axis.maxValue = (
                axis_range["min"],
                axis_range["min"],
                axis_range["max"],
            )
            # TODO: get proper name string of axis
            axis.axisNameID = nameTable.addMultilingualName(
                {"en": axis.axisTag}, font, minNameID=256, mac=False
            )
            font["fvar"].axes.append(axis)

    # Following code from:
    # https://github.com/googlefonts/generate-avar2/blob/a8625dc6802420d2195c7475477828c50ebc13c0/RobotoFlex-avar2.py#L311-L327
    # Thanks Behdad!
    print("Update various VarStores")
    stores = []
    if (
        "GDEF" in font
        and hasattr(font["GDEF"], "table")
        and hasattr(font["GDEF"].table, "VarStore")
    ):
        stores.append(font["GDEF"].table.VarStore)
    if (
        "HVAR" in font
        and hasattr(font["HVAR"], "table")
        and hasattr(font["HVAR"].table, "VarStore")
    ):
        stores.append(font["HVAR"].table.VarStore)
    if (
        "MVAR" in font
        and hasattr(font["MVAR"], "table")
        and hasattr(font["MVAR"].table, "VarStore")
    ):
        stores.append(font["MVAR"].table.VarStore)
    nullRegion = otTables.VarRegionAxis()
    nullRegion.StartCoord = -1
    nullRegion.PeakCoord = 0
    nullRegion.EndCoord = 1
    for store in stores:
        store.VarRegionList.RegionAxisCount = len(font["fvar"].axes)
        for region in store.VarRegionList.Region:
            while len(region.VarRegionAxis) < len(font["fvar"].axes):
                region.VarRegionAxis.append(nullRegion)


def gen_avar2_mapping(font, mapping):
    gen_fvar_axes(font, mapping)
    axisTags = [axis.axisTag for axis in font["fvar"].axes]
    axis_tag_to_name = {}
    nametable = font["name"]
    axes = OrderedDict()
    for axis in font["fvar"].axes:
        if axis.axisTag in axis_tag_to_name:  # Internal axis, ignore
            continue
        axis_name = nametable.getName(axis.axisNameID, 3, 1, 0x409).toUnicode()
        axis_tag_to_name[axis.axisTag] = axis_name

        if "avar" in font:
            axis_map = [
                (_denormalize(k, axis), _denormalize(v, axis))
                for k, v in font["avar"].segments[axis.axisTag].items()
            ]
            axis_map = list(sorted(set(axis_map)))
        else:
            axis_map = []
        axes[axis_name] = AxisDescriptor(
            tag=axis.axisTag,
            name=axis_name,
            minimum=axis.minValue,
            maximum=axis.maxValue,
            default=axis.defaultValue,
            map=axis_map,
            axisOrdering=None,
            axisLabels=[],
        )

    mapping = [
        AxisMappingDescriptor(
            inputLocation={axis_tag_to_name[k]: v for k, v in m["in"].items()},
            outputLocation={axis_tag_to_name[k]: v for k, v in m["out"].items()},
        )
        for m in mapping
    ]

    if "avar" in font:
        del font["avar"]
    _add_avar(font, axes, mapping, axisTags)

    # Add new axes to fvar instances
    for inst in font["fvar"].instances:
        for axis in font["fvar"].axes:
            if axis.axisTag not in inst.coordinates:
                inst.coordinates[axis.axisTag] = axis.defaultValue


def load_fontra(f):
    """Load a Fontra font-data.json file and extract the axis mapping."""
    import json

    data = json.load(f)
    if not isinstance(data, dict) or "fontInfo" not in data:
        raise ValueError("JSON file must be a Fontra font-data.json")
    axes = data["axes"]["axes"]
    mapping = data["axes"].get("mappings", [])
    # We want: list of { 'in' : { tag: val, tag: val}, 'out': { tag: val, tag: val} }
    # We have: list of { "inputLocation": { name: val, name: val}, "outputLocation": { name: val, name: val} }
    # Let's first map axis names to tags
    name_to_tag = {axis["name"]: axis["tag"] for axis in axes}
    new_map = [
        {
            "in": {name_to_tag[k]: v for k, v in m["inputLocation"].items()},
            "out": {name_to_tag[k]: v for k, v in m["outputLocation"].items()},
        }
        for m in mapping
    ]
    # Check all input locations are unique
    seen_locations = set()
    for m in new_map:
        if tuple(m["in"].items()) in seen_locations:
            raise ValueError(
                f"Duplicate input location in mapping: {m['in']}. Input locations must be unique."
            )
        seen_locations.add(tuple(m["in"].items()))
    return new_map


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "fonts", nargs="+", help="Variable TTF files which make up a family"
    )
    parser.add_argument("src", help="use yaml/json file to build STAT", default=None)
    out_group = parser.add_mutually_exclusive_group(required=True)
    out_group.add_argument("--out", "-o", help="Output dir for fonts")
    out_group.add_argument(
        "--inplace", action="store_true", default=False, help="Overwrite input files"
    )
    args = parser.parse_args(args)

    fonts = [TTFont(f) for f in args.fonts]
    if args.src.endswith(".json"):
        config = load_fontra(open(args.src))
        for font in fonts:
            gen_avar2_mapping(font, config)
    else:
        config = yaml.load(open(args.src), Loader=yaml.SafeLoader)
        for font in fonts:
            filename = os.path.basename(font.reader.file.name)
            gen_avar2_mapping(font, config[filename])

    if args.inplace:
        for font in fonts:
            font.save(font.reader.file.name)
    elif args.out:
        if not os.path.isdir(args.out):
            os.mkdir(args.out)
        for font in fonts:
            font.save(os.path.join(args.out, os.path.basename(font.reader.file.name)))


if __name__ == "__main__":
    main()



from gftools.builder.operations import FontmakeOperationBase


class BuildVariable(FontmakeOperationBase):
    description = "Build a variable font from a source file"
    rule = "fontmake --output-path $out -o variable $fontmake_type $in $args"
